from __future__ import annotations

from typing import Tuple

from amulet.api.history import Changeable
from amulet.api.data_types import Dimension

LOCAL_PLAYER = "~local_player"


class Player(Changeable):
    def __init__(
        self,
        player_id: str,
        dimension: str,
        location: Tuple[float, float, float],
        rotation: Tuple[float, float],
    ):
        """
        Creates a new instance of :class:`Player` with the given UUID, location, and rotation

        :param player_id: The ID of the player
        :param location: The location of the player in world coordinates
        :param rotation: The rotation of the player
        :param dimension: The dimension the player is in
        """
        super().__init__()
        assert isinstance(player_id, str)
        assert (
            isinstance(location, tuple)
            and len(location) == 3
            and all(isinstance(f, float) for f in location)
        )
        assert (
            isinstance(rotation, tuple)
            and len(rotation) == 2
            and all(isinstance(f, float) for f in rotation)
        )
        self._player_id = player_id
        self._location = location
        self._rotation = rotation
        self._dimension = dimension

    def __repr__(self):
        return f"Player({self.player_id}, {self.dimension}, {self.location}, {self.rotation})"

    @property
    def player_id(self) -> str:
        """The player's ID"""
        return self._player_id

    @property
    def location(self) -> Tuple[float, float, float]:
        """The current location of the player in the world"""
        return self._location

    @property
    def rotation(self) -> Tuple[float, float]:
        """The current rotation of the player in the world"""
        return self._rotation

    @property
    def dimension(self) -> Dimension:
        """The current dimension the player is in"""
        return self._dimension



from typing import Any, Union, Generator, Callable, Tuple, TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from amulet.api.level import BaseLevel
    from .world_types import Dimension

# Operation types

#: The data type that an operation is able to yield.
OperationYieldType = Union[int, float, Tuple[Union[int, float], str]]

#: The data type that an operation is able to return.
OperationReturnType = Optional[Union[Generator[OperationYieldType, None, Any], Any]]

#: The data type for an operation callable object.
OperationType = Callable[["BaseLevel", "Dimension", Any], OperationReturnType]



from __future__ import annotations

import itertools
import numpy
import math

from typing import Tuple, Iterable, Generator, Optional, Union

from amulet.api.data_types import (
    BlockCoordinates,
    BlockCoordinatesAny,
    CoordinatesAny,
    ChunkCoordinates,
    SubChunkCoordinates,
    FloatTriplet,
    PointCoordinatesAny,
)
from amulet.utils.world_utils import (
    block_coords_to_chunk_coords,
    blocks_slice_to_chunk_slice,
)
from amulet.utils.matrix import (
    transform_matrix,
    displacement_matrix,
)
from .abstract_selection import AbstractBaseSelection
from .. import selection


class SelectionBox(AbstractBaseSelection):
    """
    The SelectionBox class represents a single cuboid selection.

    When combined with :class:`~amulet.api.selection.SelectionGroup` it can represent any arbitrary shape.
    """

    __slots__ = (
        "_min_x",
        "_min_y",
        "_min_z",
        "_max_x",
        "_max_y",
        "_max_z",
        "_point_1",
        "_point_2",
    )

    def __init__(self, point_1: BlockCoordinatesAny, point_2: BlockCoordinatesAny):
        """
        Construct a new SelectionBox instance.

        >>> # a selection box that selects one block.
        >>> box = SelectionBox(
        >>>     (0, 0, 0),
        >>>     (1, 1, 1)
        >>> )

        :param point_1: The first point of the selection.
        :param point_2: The second point of the selection.
        """
        box = numpy.array([point_1, point_2]).round().astype(int)
        p1, p2 = box.tolist()
        self._point_1 = tuple(p1)
        self._point_2 = tuple(p2)
        self._min_x, self._min_y, self._min_z = numpy.min(box, 0).tolist()
        self._max_x, self._max_y, self._max_z = numpy.max(box, 0).tolist()

    @classmethod
    def create_chunk_box(
        cls, cx: int, cz: int, sub_chunk_size: int = 16
    ) -> SelectionBox:
        """
        Get a :class:`SelectionBox` containing the whole of a given chunk.

        >>> box = SelectionBox.create_chunk_box(1, 2)
        SelectionBox((16, -1073741824, 32), (32, 1073741824, 48))

        :param cx: The x coordinate of the chunk
        :param cz: The z coordinate of the chunk
        :param sub_chunk_size: The dimension of a sub-chunk. Default 16.
        """
        return cls(
            (cx * sub_chunk_size, -(2**30), cz * sub_chunk_size),
            ((cx + 1) * sub_chunk_size, 2**30, (cz + 1) * sub_chunk_size),
        )

    @classmethod
    def create_sub_chunk_box(
        cls, cx: int, cy: int, cz: int, sub_chunk_size: int = 16
    ) -> SelectionBox:
        """
        Get a :class:`SelectionBox` containing the whole of a given sub-chunk.

        >>> SelectionBox.create_sub_chunk_box(1, 0, 2)
        SelectionBox((16, 0, 32), (32, 16, 48))

        :param cx: The x coordinate of the chunk
        :param cy: The y coordinate of the chunk
        :param cz: The z coordinate of the chunk
        :param sub_chunk_size: The dimension of a sub-chunk. Default 16.
        """
        return cls(
            (cx * sub_chunk_size, cy * sub_chunk_size, cz * sub_chunk_size),
            (
                (cx + 1) * sub_chunk_size,
                (cy + 1) * sub_chunk_size,
                (cz + 1) * sub_chunk_size,
            ),
        )

    def create_moved_box(
        self, offset: BlockCoordinatesAny, subtract=False
    ) -> SelectionBox:
        """
        Create a new :class:`SelectionBox` based on this one with the coordinates moved by the given offset.

        :param offset: The amount to move the box.
        :param subtract: If true will subtract the offset rather than adding.
        :return: The new selection with the given offset.
        """
        offset = numpy.array(offset)
        if subtract:
            offset *= -1
        return SelectionBox(offset + self.min, offset + self.max)

    def chunk_locations(self, sub_chunk_size: int = 16) -> Iterable[ChunkCoordinates]:
        cx_min, cz_min, cx_max, cz_max = block_coords_to_chunk_coords(
            self.min_x,
            self.min_z,
            self.max_x - 1,
            self.max_z - 1,
            sub_chunk_size=sub_chunk_size,
        )
        yield from itertools.product(
            range(cx_min, cx_max + 1), range(cz_min, cz_max + 1)
        )

    def chunk_boxes(
        self, sub_chunk_size: int = 16
    ) -> Iterable[Tuple[ChunkCoordinates, SelectionBox]]:
        for cx, cz in self.chunk_locations(sub_chunk_size):
            yield (cx, cz), self.intersection(
                SelectionBox.create_chunk_box(cx, cz, sub_chunk_size)
            )

    def chunk_y_locations(self, sub_chunk_size: int = 16) -> Iterable[int]:
        """
        An iterable of all the sub-chunk y indexes this box intersects.

        :param sub_chunk_size: The dimension of a sub-chunk. Default 16.
        """
        cy_min, cy_max = block_coords_to_chunk_coords(
            self.min_y, self._max_y - 1, sub_chunk_size=sub_chunk_size
        )
        for cy in range(cy_min, cy_max + 1):
            yield cy

    def sub_chunk_locations(
        self, sub_chunk_size: int = 16
    ) -> Iterable[SubChunkCoordinates]:
        for cx, cz in self.chunk_locations(sub_chunk_size):
            for cy in self.chunk_y_locations(sub_chunk_size):
                yield cx, cy, cz

    def chunk_count(self, sub_chunk_size: int = 16) -> int:
        cx_min, cz_min, cx_max, cz_max = block_coords_to_chunk_coords(
            self.min_x,
            self.min_z,
            self.max_x - 1,
            self.max_z - 1,
            sub_chunk_size=sub_chunk_size,
        )
        return (cx_max + 1 - cx_min) * (cz_max + 1 - cz_min)

    def sub_chunk_count(self, sub_chunk_size: int = 16) -> int:
        cy_min, cy_max = block_coords_to_chunk_coords(
            self.min_y,
            self.max_y - 1,
            sub_chunk_size=sub_chunk_size,
        )
        return (cy_max + 1 - cy_min) * self.chunk_count()

    def sub_chunk_boxes(
        self, sub_chunk_size: int = 16
    ) -> Iterable[Tuple[SubChunkCoordinates, SelectionBox]]:
        for cx, cy, cz in self.sub_chunk_locations(sub_chunk_size):
            yield (cx, cy, cz), self.intersection(
                SelectionBox.create_sub_chunk_box(cx, cy, cz, sub_chunk_size)
            )

    def __iter__(self) -> Iterable[BlockCoordinates]:
        """An iterable of all the block locations within this box."""
        return self.blocks

    @property
    def blocks(self) -> Iterable[BlockCoordinates]:
        return itertools.product(
            range(self._min_x, self._max_x),
            range(self._min_y, self._max_y),
            range(self._min_z, self._max_z),
        )

    def __repr__(self) -> str:
        return f"SelectionBox({self.point_1}, {self.point_2})"

    def __str__(self) -> str:
        return f"({self.point_1}, {self.point_2})"

    def __contains__(self, item: CoordinatesAny) -> bool:
        return self.contains_block(item)

    def contains_block(self, coords: CoordinatesAny) -> bool:
        return (
            self._min_x <= coords[0] < self._max_x
            and self._min_y <= coords[1] < self._max_y
            and self._min_z <= coords[2] < self._max_z
        )

    def contains_point(self, coords: CoordinatesAny) -> bool:
        return (
            self._min_x <= coords[0] <= self._max_x
            and self._min_y <= coords[1] <= self._max_y
            and self._min_z <= coords[2] <= self._max_z
        )

    def __eq__(self, other) -> bool:
        return self.min == other.min and self.max == other.max

    def __ne__(self, other) -> bool:
        return not self == other

    def __hash__(self) -> int:
        return hash((*self.min, *self.max))

    @property
    def slice(self) -> Tuple[slice, slice, slice]:
        """
        Converts the :class:`SelectionBox` minimum/maximum coordinates into slice arguments

        :return: The :class:`SelectionBox` coordinates as slices in (x,y,z) order
        """
        return (
            slice(self._min_x, self._max_x),
            slice(self._min_y, self._max_y),
            slice(self._min_z, self._max_z),
        )

    def chunk_slice(
        self, cx: int, cz: int, sub_chunk_size: int = 16
    ) -> Tuple[slice, slice, slice]:
        """
        Get the slice of the box in relative form for a given chunk.

        >>> SelectionBox((0, 0, 0), (32, 32, 32)).chunk_slice(1, 1)
        (slice(0, 16, None), slice(0, 32, None), slice(0, 16, None))

        :param cx: The x coordinate of the chunk
        :param cz: The z coordinate of the chunk
        :param sub_chunk_size: The dimension of a sub-chunk. Default 16.
        """
        s_x, s_y, s_z = self.slice
        x_chunk_slice = blocks_slice_to_chunk_slice(s_x, sub_chunk_size, cx)
        z_chunk_slice = blocks_slice_to_chunk_slice(s_z, sub_chunk_size, cz)
        return x_chunk_slice, s_y, z_chunk_slice

    def sub_chunk_slice(
        self, cx: int, cy: int, cz: int, sub_chunk_size: int = 16
    ) -> Tuple[slice, slice, slice]:
        """
        Get the slice of the box in relative form for a given sub-chunk.

        >>> SelectionBox((0, 0, 0), (32, 32, 32)).sub_chunk_slice(1, 1, 1)
        (slice(0, 16, None), slice(0, 16, None), slice(0, 16, None))

        :param cx: The x coordinate of the chunk
        :param cy: The y coordinate of the chunk
        :param cz: The z coordinate of the chunk
        :param sub_chunk_size: The dimension of a sub-chunk. Default 16.
        """
        x_chunk_slice, s_y, z_chunk_slice = self.chunk_slice(cx, cz, sub_chunk_size)
        y_chunk_slice = blocks_slice_to_chunk_slice(s_y, sub_chunk_size, cy)
        return x_chunk_slice, y_chunk_slice, z_chunk_slice

    @property
    def point_1(self) -> BlockCoordinates:
        """The first value given to the constructor."""
        return self._point_1

    @property
    def point_2(self) -> BlockCoordinates:
        """The second value given to the constructor."""
        return self._point_2

    @property
    def points(self) -> Tuple[BlockCoordinates, BlockCoordinates]:
        """The points given to the constructor."""
        return self.point_1, self.point_2

    @property
    def points_array(self) -> numpy.ndarray:
        """The points given to the constructor as a numpy array."""
        return numpy.array(self.points)

    @property
    def min_x(self) -> int:
        return self._min_x

    @property
    def min_y(self) -> int:
        return self._min_y

    @property
    def min_z(self) -> int:
        return self._min_z

    @property
    def max_x(self) -> int:
        return self._max_x

    @property
    def max_y(self) -> int:
        return self._max_y

    @property
    def max_z(self) -> int:
        return self._max_z

    @property
    def min(self) -> BlockCoordinates:
        return self._min_x, self._min_y, self._min_z

    @property
    def min_array(self) -> numpy.ndarray:
        return numpy.array(self.min)

    @property
    def max(self) -> BlockCoordinates:
        return self._max_x, self._max_y, self._max_z

    @property
    def max_array(self) -> numpy.ndarray:
        return numpy.array(self.max)

    @property
    def bounds(self) -> Tuple[BlockCoordinates, BlockCoordinates]:
        return (
            (self._min_x, self._min_y, self._min_z),
            (self._max_x, self._max_y, self._max_z),
        )

    @property
    def bounds_array(self) -> numpy.ndarray:
        return numpy.array(self.bounds)

    @property
    def size_x(self) -> int:
        """The length of the box in the x axis."""
        return self._max_x - self._min_x

    @property
    def size_y(self) -> int:
        """The length of the box in the y axis."""
        return self._max_y - self._min_y

    @property
    def size_z(self) -> int:
        """The length of the box in the z axis."""
        return self._max_z - self._min_z

    @property
    def shape(self) -> Tuple[int, int, int]:
        """
        The shape of the box.

        >>> SelectionBox((0, 0, 0), (1, 1, 1)).shape
        (1, 1, 1)
        """
        return self.size_x, self.size_y, self.size_z

    @property
    def volume(self) -> int:
        """
        The number of blocks in the box.

        >>> SelectionBox((0, 0, 0), (1, 1, 1)).shape
        1
        """
        return self.size_x * self.size_y * self.size_z

    def touches(self, other: SelectionBox) -> bool:
        """
        Method to check if this instance of :class:`SelectionBox` touches but does not intersect another SelectionBox.

        :param other: The other SelectionBox
        :return: True if the two :class:`SelectionBox` instances touch, False otherwise
        """
        # It touches if the box does not intersect but intersects when expanded by one block.
        # There may be a simpler way to do this.
        return self.touches_or_intersects(other) and not self.intersects(other)

    def touches_or_intersects(self, other: SelectionBox) -> bool:
        """
        Method to check if this instance of SelectionBox touches or intersects another SelectionBox.

        :param other: The other SelectionBox.
        :return: True if the two :class:`SelectionBox` instances touch or intersect, False otherwise.
        """
        return not (
            self.min_x >= other.max_x + 1
            or self.min_y >= other.max_y + 1
            or self.min_z >= other.max_z + 1
            or self.max_x <= other.min_x - 1
            or self.max_y <= other.min_y - 1
            or self.max_z <= other.min_z - 1
        )

    def intersects(self, other: SelectionBox) -> bool:
        """
        Method to check whether this instance of SelectionBox intersects another SelectionBox.

        :param other: The other SelectionBox to check for intersection.
        :return: True if the two :class:`SelectionBox` instances intersect, False otherwise.
        """
        return not (
            self.min_x >= other.max_x
            or self.min_y >= other.max_y
            or self.min_z >= other.max_z
            or self.max_x <= other.min_x
            or self.max_y <= other.min_y
            or self.max_z <= other.min_z
        )

    def contains_box(self, other: SelectionBox) -> bool:
        """
        Method to check if the other SelectionBox other fits entirely within this instance of SelectionBox.

        :param other: The SelectionBox to test.
        :return: True if other fits with self, False otherwise.
        """
        return (
            self.min_x <= other.min_x
            and self.min_y <= other.min_y
            and self.min_z <= other.min_z
            and other.max_x <= self.max_x
            and other.max_y <= self.max_y
            and other.max_z <= self.max_z
        )

    def intersection(self, other: SelectionBox) -> SelectionBox:
        return SelectionBox(
            numpy.clip(other.min, self.min, self.max),
            numpy.clip(other.max, self.min, self.max),
        )

    def subtract(self, other: SelectionBox) -> selection.SelectionGroup:
        """
        Get a :class:`~amulet.api.selection.SelectionGroup` containing boxes that are in self but not in other.

        This may be empty if other fully contains self or equal to self if they do not intersect.

        :param other: The SelectionBox to subtract.
        :return:
        """
        if self.intersects(other):
            other = self.intersection(other)
            if self == other:
                # if the two selections are the same there is no difference.
                return selection.SelectionGroup()
            else:
                boxes = []
                if self.min_y < other.min_y:
                    # bottom box
                    boxes.append(
                        SelectionBox(
                            (self.min_x, self.min_y, self.min_z),
                            (self.max_x, other.min_y, self.max_z),
                        )
                    )

                if other.max_y < self.max_y:
                    # top box
                    boxes.append(
                        SelectionBox(
                            (self.min_x, other.max_y, self.min_z),
                            (self.max_x, self.max_y, self.max_z),
                        )
                    )

                # BBB  NNN  TTT
                # BBB  WOE  TTT
                # BBB  SSS  TTT

                if self.min_z < other.min_z:
                    # north box
                    boxes.append(
                        SelectionBox(
                            (self.min_x, other.min_y, self.min_z),
                            (self.max_x, other.max_y, other.min_z),
                        )
                    )

                if other.max_z < self.max_z:
                    # south box
                    boxes.append(
                        SelectionBox(
                            (self.min_x, other.min_y, other.max_z),
                            (self.max_x, other.max_y, self.max_z),
                        )
                    )

                if self.min_x < other.min_x:
                    # west box
                    boxes.append(
                        SelectionBox(
                            (self.min_x, other.min_y, other.min_z),
                            (other.min_x, other.max_y, other.max_z),
                        )
                    )

                if other.max_x < self.max_x:
                    # east box
                    boxes.append(
                        SelectionBox(
                            (other.max_x, other.min_y, other.min_z),
                            (self.max_x, other.max_y, other.max_z),
                        )
                    )

                return selection.SelectionGroup(boxes)
        else:
            # if the boxes do not intersect then the difference is self
            return selection.SelectionGroup(self)

    def intersects_vector(
        self, origin: PointCoordinatesAny, vector: PointCoordinatesAny
    ) -> Optional[float]:
        """
        Determine if a vector from a given point collides with this selection box.

        :param origin: Location of the origin of the vector
        :param vector: The look vector
        :return: Multiplier of the vector to the collision location. None if it does not collide
        """
        # Logic based on https://www.scratchapixel.com/lessons/3d-basic-rendering/minimal-ray-tracer-rendering-simple-shapes/ray-box-intersection
        for obj in (origin, vector):
            if isinstance(obj, (tuple, list)):
                if len(obj) != 3 or not all(type(o) in (int, float) for o in obj):
                    raise ValueError(
                        "Given tuple/list type must contain three ints or floats."
                    )
            elif isinstance(obj, numpy.ndarray):
                if obj.shape != (3,) and numpy.issubdtype(obj.dtype, numpy.number):
                    raise ValueError(
                        "Given ndarray type must have a numerical data type with length three."
                    )
        vector = numpy.array(vector)
        vector[abs(vector) < 0.000001] = 0.000001
        (tmin, tymin, tzmin), (tmax, tymax, tzmax) = numpy.sort(
            (self.bounds_array - numpy.array(origin)) / numpy.array(vector), axis=0
        )

        if tmin > tymax or tymin > tmax:
            return None

        if tymin > tmin:
            tmin = tymin

        if tymax < tmax:
            tmax = tymax

        if tmin > tzmax or tzmin > tmax:
            return None

        if tzmin > tmin:
            tmin = tzmin

        if tzmax < tmax:
            tmax = tzmax

        if tmin >= 0:
            return tmin
        elif tmax >= 0:
            return tmax
        else:
            return None

    @staticmethod
    def _transform_points(points: numpy.ndarray, matrix: numpy.ndarray):
        assert (
            isinstance(points, numpy.ndarray)
            and len(points.shape) == 2
            and points.shape[1] == 3
        )
        assert isinstance(matrix, numpy.ndarray) and matrix.shape == (4, 4)
        points_array = numpy.ones((points.shape[0], 4))
        points_array[:, :3] = points
        return numpy.matmul(
            matrix,
            points_array.T,
        ).T[:, :3]

    def _iter_transformed_boxes(
        self, transform: numpy.ndarray
    ) -> Generator[
        Tuple[
            float,  # progress
            SelectionBox,  # The sub-chunk box.
            Union[
                numpy.ndarray,  # The bool array of which of the transformed blocks are contained.
                bool,  # If True all blocks are contained, if False no blocks are contained.
            ],
            Optional[numpy.ndarray],  # A float array of where those blocks came from.
        ],
        None,
        None,
    ]:
        """The core logic for transform and transformed_points"""
        assert isinstance(transform, numpy.ndarray) and transform.shape == (4, 4)
        inverse_transform = numpy.linalg.inv(transform)
        inverse_transform2 = numpy.linalg.inv(
            numpy.matmul(displacement_matrix(-0.5, -0.5, -0.5), transform)
        )

        def transform_box(box_: SelectionBox, transform_) -> SelectionBox:
            """transform a box and get the AABB that contains this rotated box."""

            # find the transformed points of each of the corners
            points = numpy.matmul(
                transform_,
                numpy.array(
                    list(
                        itertools.product(
                            [box_.min_x, box_.max_x],
                            [box_.min_y, box_.max_y],
                            [box_.min_z, box_.max_z],
                            [1],
                        )
                    )
                ).T,
            ).T[:, :3]
            # this is a larger AABB that contains the roatated box and a bit more.
            return SelectionBox(numpy.min(points, axis=0), numpy.max(points, axis=0))

        aabb = transform_box(self, transform)
        count = aabb.sub_chunk_count()
        index = 0

        for _, box in aabb.sub_chunk_boxes():
            index += 1
            original_box = transform_box(box, inverse_transform)
            if self.intersects(original_box):
                # if the boxes do not intersect then nothing needs doing.
                if self.contains_box(original_box):
                    # if the box is fully contained use the whole box.
                    yield index / count, box, True, None
                else:
                    # the original points the transformed locations relate to
                    original_blocks = self._transform_points(
                        numpy.transpose(
                            numpy.mgrid[
                                box.min_x : box.max_x,
                                box.min_y : box.max_y,
                                box.min_z : box.max_z,
                            ],
                            (1, 2, 3, 0),
                        ).reshape(-1, 3),
                        inverse_transform2,
                    )

                    box_shape = box.shape
                    mask: numpy.ndarray = numpy.all(
                        numpy.logical_and(
                            original_blocks < self.max, original_blocks >= self.min
                        ),
                        axis=1,
                    ).reshape(box_shape)

                    yield index / count, box, mask, original_blocks.reshape(
                        box_shape + (3,)
                    )
            else:
                yield index / count, box, False, None

    def transformed_points(
        self, transform: numpy.ndarray
    ) -> Iterable[Tuple[float, Optional[numpy.ndarray], Optional[numpy.ndarray]]]:
        """
        Get the locations of the transformed blocks and the source blocks they came from.

        :param transform: The matrix that this box will be transformed by.
        :return: An iterable of two Nx3 numpy arrays of the source block locations and the destination block locations. The destination locations will be unique but the source may not be and some may not be included.
        """
        for progress, box, mask, original in self._iter_transformed_boxes(transform):
            if isinstance(mask, bool) and mask:
                new_points = numpy.transpose(
                    numpy.mgrid[
                        box.min_x : box.max_x,
                        box.min_y : box.max_y,
                        box.min_z : box.max_z,
                    ],
                    (1, 2, 3, 0),
                ).reshape(-1, 3)
                old_points = self._transform_points(
                    new_points,
                    numpy.linalg.inv(
                        numpy.matmul(displacement_matrix(-0.5, -0.5, -0.5), transform)
                    ),
                )
                yield progress, old_points, new_points
            elif isinstance(mask, numpy.ndarray) and numpy.any(mask):
                yield progress, original[mask], box.min_array + numpy.argwhere(mask)
            else:
                yield progress, None, None

    def transform(
        self, scale: FloatTriplet, rotation: FloatTriplet, translation: FloatTriplet
    ) -> selection.SelectionGroup:
        """
        Creates a :class:`~amulet.api.selection.SelectionGroup` of transformed SelectionBox(es).

        :param scale: A tuple of scaling factors in the x, y and z axis.
        :param rotation: The rotation about the x, y and z axis in radians.
        :param translation: The translation about the x, y and z axis.
        :return: A new :class:`~amulet.api.selection.SelectionGroup` representing the transformed selection.
        """
        quadrant = math.pi / 2
        if all(abs(r - quadrant * round(r / quadrant)) < 0.0001 for r in rotation):
            min_point, max_point = numpy.matmul(
                transform_matrix(scale, rotation, translation),
                numpy.array([[*self.min, 1], [*self.max, 1]]).T,
            ).T[:, :3]
            return selection.SelectionGroup(SelectionBox(min_point, max_point))
        else:
            boxes = []
            for _, box, mask, _ in self._iter_transformed_boxes(
                transform_matrix(scale, rotation, translation)
            ):
                if isinstance(mask, bool):
                    if mask:
                        boxes.append(box)
                else:
                    box_shape = box.shape
                    any_array: numpy.ndarray = numpy.any(mask, axis=2)
                    box_2d_shape = numpy.array(any_array.shape)
                    any_array_flat = any_array.ravel()
                    start_array = numpy.argmax(mask, axis=2)
                    stop_array = box_shape[2] - numpy.argmax(
                        numpy.flip(mask, axis=2), axis=2
                    )
                    # effectively a greedy meshing algorithm in 2D
                    index = 0
                    while index < any_array_flat.size:
                        # while there are unhandled true values
                        index = numpy.argmax(any_array_flat[index:]) + index
                        # find the first true value
                        if any_array_flat[index]:
                            # check that that value is actually True
                            # create the bounds for the box
                            min_x, min_y = max_x, max_y = numpy.unravel_index(
                                index, box_2d_shape
                            )
                            # find the z bounds
                            min_z = start_array[min_x, min_y]
                            max_z = stop_array[min_x, min_y]
                            while max_x < box_2d_shape[0] - 1:
                                # expand in the x while the bounds are the same
                                new_max_x = max_x + 1
                                if (
                                    any_array[new_max_x, max_y]
                                    and start_array[new_max_x, max_y] == min_z
                                    and stop_array[new_max_x, max_y] == max_z
                                ):
                                    # the box z values are the same
                                    max_x = new_max_x
                                else:
                                    break
                            while max_y < box_2d_shape[1] - 1:
                                # expand in the y while the bounds are the same
                                new_max_y = max_y + 1
                                if (
                                    numpy.all(any_array[min_x : max_x + 1, new_max_y])
                                    and numpy.all(
                                        start_array[min_x : max_x + 1, new_max_y]
                                        == min_z
                                    )
                                    and numpy.all(
                                        stop_array[min_x : max_x + 1, new_max_y]
                                        == max_z
                                    )
                                ):
                                    # the box z values are the same
                                    max_y = new_max_y
                                else:
                                    break
                            boxes.append(
                                SelectionBox(
                                    box.min_array + (min_x, min_y, min_z),
                                    box.min_array + (max_x + 1, max_y + 1, max_z),
                                )
                            )
                            any_array[min_x : max_x + 1, min_y : max_y + 1] = False
                        else:
                            # If there are no more True values argmax will return 0
                            break
            return selection.SelectionGroup(boxes)



from __future__ import annotations

from .leveldb_38 import (
    LevelDB38Interface as ParentInterface,
)


class LevelDB39Interface(ParentInterface):
    chunk_version = 39


export = LevelDB39Interface



from typing import Union, Tuple
import numpy

# from nptyping import NDArray

#: The data type for a Tuple containing three floats. Use :data:`PointCoordinates` for x, y, z float coordinates.
FloatTriplet = Tuple[float, float, float]

#: The data type for the dimension identifier.
Dimension = str

#: The data type for the x and z coordinates of a region file. Note these are no the same as block coordinates.
RegionCoordinates = Tuple[int, int]

#: The data type for the x and z coordinate of a chunk. Note these are no the same as block coordinates.
ChunkCoordinates = Tuple[int, int]

#: The data type for the x, y and z coordinates of a sub-chunk within the world. Note these are no the same as block coordinates.
SubChunkCoordinates = Tuple[int, int, int]

#: The data type for the dimension and x and z coordinates of a chunk within a world. Note these are no the same as block coordinates.
DimensionCoordinates = Tuple[Dimension, int, int]

#: The data type for the x, y and z location of a block within the world.
BlockCoordinates = Tuple[int, int, int]

#: The data type for the x, y and z location of a block within the world in the form of a numpy array.
BlockCoordinatesNDArray = numpy.ndarray  # NDArray[(3, ), int]

#: The data type for the x, y and z location of a block within the world in either tuple or numpy array form.
BlockCoordinatesAny = Union[BlockCoordinates, BlockCoordinatesNDArray]

#: The data type for the x, y and z location of a point within the world.
PointCoordinates = FloatTriplet

#: The data type for the x, y and z location of a point within the world in the form of a numpy array.
PointCoordinatesNDArray = numpy.ndarray  # NDArray[(3, ), numpy.float]

#: The data type for the x, y and z location of a point within the world in either tuple or numpy array form.
PointCoordinatesAny = Union[PointCoordinates, PointCoordinatesNDArray]

#: The data type for x, y and z location in either float or int and tuple or numpy array format.
CoordinatesAny = Union[BlockCoordinatesAny, PointCoordinatesAny]

#: The data type for a sub-chunk array. This array should be a 16x16x16 numpy array.
SubChunkNDArray = numpy.ndarray  # NDArray[(16, 16, 16), numpy.uint32]

#: The data type for a biome string identifier.
BiomeType = str



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Iterable, Tuple
import numpy
from .. import selection

from amulet.api.data_types import (
    CoordinatesAny,
    ChunkCoordinates,
    SubChunkCoordinates,
    FloatTriplet,
    BlockCoordinates,
)


class AbstractBaseSelection(ABC):
    """
    A parent selection class to force a consistent API for selection group and box.
    """

    __slots__ = ()

    @abstractmethod
    def __contains__(self, item: CoordinatesAny) -> bool:
        """
        Is the given block/point contained within the selection.

        >>> (1, 2, 3) in selection1
        True
        >>> (1.5, 2.5, 3.5) in selection1
        True

        Will return False if the coordinate is on the most positive edge.
        See :meth:`contains_point` to include the top edge.

        :param item: The block/point to check.
        :return: True if the block/point is in the selection, False otherwise.
        """
        raise NotImplementedError

    @abstractmethod
    def __eq__(self, other) -> bool:
        """
        Is the selection equal to the other selection.

        >>> selection1 == selection2
        True

        :param other: The other selection to test.
        :return: True if the selections are the same, False otherwise.
        """
        raise NotImplementedError

    @abstractmethod
    def __iter__(self):
        raise NotImplementedError

    @abstractmethod
    def __repr__(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def __str__(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def blocks(self) -> Iterable[BlockCoordinates]:
        """
        The location of every block in the selection.

        :return: An iterable of block locations.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def bounds(self) -> Tuple[BlockCoordinates, BlockCoordinates]:
        """
        The minimum and maximum points in the selection.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def bounds_array(self) -> numpy.ndarray:
        """
        The minimum and maximum points in the selection as a numpy array.
        """
        raise NotImplementedError

    @abstractmethod
    def chunk_boxes(
        self, sub_chunk_size: int = 16
    ) -> Iterable[Tuple[ChunkCoordinates, selection.box.SelectionBox]]:
        """
        An iterable of chunk coordinates and boxes that intersect the selection and the chunk.

        >>> for (cx, cz), box in selection1.chunk_boxes():
        >>>     ...

        :param sub_chunk_size: The dimension of a sub-chunk. Default 16.
        """
        raise NotImplementedError

    @abstractmethod
    def chunk_count(self, sub_chunk_size: int = 16) -> int:
        """
        The number of chunks that intersect the selection.

        :param sub_chunk_size: The dimension of a sub-chunk. Default 16.
        """
        raise NotImplementedError

    @abstractmethod
    def chunk_locations(self, sub_chunk_size: int = 16) -> Iterable[ChunkCoordinates]:
        """
        An iterable of chunk coordinates that intersect the selection.

        >>> for cx, cz in selection1.chunk_locations():
        >>>     ...

        :param sub_chunk_size:  The dimension of a sub-chunk. Default 16.
        """
        raise NotImplementedError

    @abstractmethod
    def contains_block(self, coords: CoordinatesAny) -> bool:
        """
        Is the block contained within the selection.

        >>> (1, 2, 3) in selection1
        True

        :param coords: The coordinate of the block defined by the most negative corner.
        :return: True if the block is in the selection.
        """
        raise NotImplementedError

    # @abstractmethod
    # def contains_box(self, other: box.SelectionBox) -> bool:
    #     raise NotImplementedError

    @abstractmethod
    def contains_point(self, coords: CoordinatesAny) -> bool:
        """
        Is the point contained within the selection.

        >>> (1.5, 2.5, 3.5) in selection1
        True

        :param coords: The coordinate of the point.
        :return: True if the point is in the selection.
        """
        raise NotImplementedError

    @abstractmethod
    def intersection(self, other):
        """
        Create and return a new a selection containing the volume that this selection and other both contain.

        If the selections do not intersect the returned selection will have no volume.

        Use :meth:`intersects` to test if the two selections intersect.

        :param other: The other selection.
        :return: A new selection containing the intersection.
        """
        raise NotImplementedError

    @abstractmethod
    def intersects(self, other) -> bool:
        """
        Does this selection intersect ``other``.

        :param other: The other selection.
        :return: True if the selections intersect, False otherwise.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def max(self) -> BlockCoordinates:
        """
        The maximum point in the selection.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def max_array(self) -> numpy.ndarray:
        """
        The maximum point in the selection as a numpy array.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def max_x(self) -> int:
        """
        The maximum x coordinate of the selection.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def max_y(self) -> int:
        """
        The maximum y coordinate of the selection.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def max_z(self) -> int:
        """
        The maximum z coordinate of the selection.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def min(self) -> BlockCoordinates:
        """
        The minimum point in the selection.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def min_array(self) -> numpy.ndarray:
        """
        The minimum point in the selection as a numpy array.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def min_x(self) -> int:
        """
        The minimum x coordinate of the selection.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def min_y(self) -> int:
        """
        The minimum y coordinate of the selection.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def min_z(self) -> int:
        """
        The minimum z coordinate of the selection.
        """
        raise NotImplementedError

    @abstractmethod
    def sub_chunk_boxes(
        self, sub_chunk_size: int = 16
    ) -> Iterable[Tuple[SubChunkCoordinates, selection.box.SelectionBox]]:
        """
        An iterable of sub-chunk coordinates and boxes that intersect the selection and the sub-chunk.

        >>> for (cx, cy, cz), box in selection1.sub_chunk_boxes():
        >>>     ...

        :param sub_chunk_size: The dimension of a sub-chunk. Default 16.
        """
        raise NotImplementedError

    @abstractmethod
    def sub_chunk_count(self, sub_chunk_size: int = 16) -> int:
        """
        The number of sub-chunks that intersect the selection.

        :param sub_chunk_size: The dimension of a sub-chunk. Default 16.
        """
        raise NotImplementedError

    @abstractmethod
    def sub_chunk_locations(
        self, sub_chunk_size: int = 16
    ) -> Iterable[SubChunkCoordinates]:
        """
        An iterable of sub-chunk coordinates that intersect the selection.

        >>> for cx, cy, cz in selection1.sub_chunk_locations():
        >>>     ...

        :param sub_chunk_size: The dimension of a sub-chunk. Default 16.
        """
        raise NotImplementedError

    @abstractmethod
    def subtract(self, other):
        raise NotImplementedError

    @abstractmethod
    def transform(
        self, scale: FloatTriplet, rotation: FloatTriplet, translation: FloatTriplet
    ):
        raise NotImplementedError

    @property
    @abstractmethod
    def volume(self) -> int:
        """
        The number of blocks in the selection.
        """
        raise NotImplementedError



import warnings

warnings.warn(
    "`amulet.libs.leveldb` is depreciated and will be removed in the future. It is replaced by the `leveldb` package.",
    DeprecationWarning,
)
from leveldb import LevelDB, LevelDBException, LevelDBEncrypted



class FormatError(Exception):
    """A base error for all errors related to the :class:`~amulet.api.wrapper.format_wrapper.FormatWrapper` class."""

    pass


class LoaderNoneMatched(FormatError):
    """An error thrown if no loader could be found that could load the given data."""

    pass


class BlockException(Exception):
    """An error thrown by the :class:`~amulet.api.block.Block` class."""

    pass


class EntryLoadError(Exception):
    pass


class EntryDoesNotExist(EntryLoadError):
    pass


class PlayerLoadError(EntryLoadError):
    """
    An error thrown if a player failed to load for some reason.
    """

    pass


class PlayerDoesNotExist(EntryDoesNotExist, PlayerLoadError):
    """
    An error thrown if a player does not exist.
    """

    pass


class ChunkLoadError(EntryLoadError):
    """
    An error thrown if a chunk failed to load for some reason.

    This may be due to a corrupt chunk, an unsupported chunk format or just because the chunk does not exist to be loaded.

    Catching this error will also catch :class:`ChunkDoesNotExist`

    >>> try:
    >>>     # get chunk
    >>>     chunk = world.get_chunk(cx, cz, dimension)
    >>> except ChunkLoadError:
    >>>     # will catch all chunks that have failed to load
    >>>     # either because they do not exist or errored during loading.
    """

    pass


class ChunkDoesNotExist(EntryDoesNotExist, ChunkLoadError):
    """
    An error thrown if a chunk does not exist and therefor cannot be loaded.

    >>> try:
    >>>     # get chunk
    >>>     chunk = world.get_chunk(cx, cz, dimension)
    >>> except ChunkDoesNotExist:
    >>>     # will catch all chunks that do not exist
    >>>     # will not catch corrupt chunks
    >>> except ChunkLoadError:
    >>>     # will only catch chunks that errored during loading
    >>>     # chunks that do not exist were caught by the previous except section.
    """

    pass


class ChunkSaveError(Exception):
    """An error thrown if there was an error during the chunk saving process."""

    pass


class DimensionDoesNotExist(Exception):
    """An error thrown if trying to load data from a dimension that does not exist."""

    pass


class ObjectReadWriteError(Exception):
    """
    An error thrown when the raw level data cannot be read from or written to.

    This is usually because the data has been opened somewhere else.
    """

    pass


class ObjectReadError(ObjectReadWriteError):
    """
    An error thrown when the raw level data cannot be read from.

    This is usually because the data has been opened somewhere else.
    """

    pass


class ObjectWriteError(ObjectReadWriteError):
    """
    An error thrown when the raw level data cannot be written to.

    This is usually because the data has been opened somewhere else.
    """

    pass



from typing import TYPE_CHECKING, Any, Tuple, Optional, List
import numpy

from amulet_nbt import IntTag, StringTag, CompoundTag

from amulet.api.wrapper import Interface, EntityIDType, EntityCoordType
from amulet.api.chunk import Chunk
from amulet.api.selection import SelectionBox
from amulet.api.data_types import (
    AnyNDArray,
    VersionNumberTuple,
    VersionIdentifierType,
    VersionNumberAny,
    PlatformType,
)
from amulet.level.loader import Translators
from amulet.api.block import Block
from .chunk import MCStructureChunk

from amulet_nbt import NamedTag

if TYPE_CHECKING:
    from amulet.api.wrapper import Translator
    from PyMCTranslate import TranslationManager


class MCStructureInterface(Interface):
    _entity_id_type = EntityIDType.namespace_str_identifier
    _entity_coord_type = EntityCoordType.Pos_list_float
    _block_entity_id_type = EntityIDType.str_id
    _block_entity_coord_type = EntityCoordType.xyz_int

    def is_valid(self, key: Tuple) -> bool:
        return True

    def decode(
        self, cx: int, cz: int, data: MCStructureChunk
    ) -> Tuple["Chunk", AnyNDArray]:
        """
        Create an amulet.api.chunk.Chunk object from raw data given by the format
        :param cx: chunk x coordinate
        :param cz: chunk z coordinate
        :param data: Raw chunk data provided by the format.
        :return: Chunk object in version-specific format, along with the block_palette for that chunk.
        """
        palette = numpy.empty(len(data.palette) + 1, dtype=object)
        palette[0] = (
            (
                17563649,
                Block(
                    namespace="minecraft",
                    base_name="air",
                    properties={"block_data": IntTag(0)},
                ),
            ),
        )

        for index, blocks in enumerate(data.palette):
            block_layers: List[Tuple[Optional[int], Block]] = []
            for block in blocks:
                *namespace_, base_name = block["name"].py_str.split(":", 1)
                namespace = namespace_[0] if namespace_ else "minecraft"
                if "version" in block:
                    version: Optional[int] = block["version"].py_int
                else:
                    version = None

                if "states" in block:  # 1.13 format
                    properties = block["states"].py_dict
                    if version is None:
                        version = 17694720  # 1, 14, 0, 0
                else:
                    properties = {"block_data": IntTag(block["val"].py_int)}
                block_layers.append(
                    (
                        version,
                        Block(
                            namespace=namespace,
                            base_name=base_name,
                            properties=properties,
                        ),
                    )
                )
            palette[index + 1] = block_layers

        chunk = Chunk(cx, cz)
        box = data.selection.create_moved_box((cx * 16, 0, cz * 16), subtract=True)
        chunk.blocks[box.slice] = data.blocks + 1
        for tag in data.block_entities:
            block_entity = self._decode_block_entity(
                NamedTag(tag), self._block_entity_id_type, self._block_entity_coord_type
            )
            if block_entity is not None:
                chunk.block_entities.insert(block_entity)
        for tag in data.entities:
            entity = self._decode_entity(
                NamedTag(tag), self._block_entity_id_type, self._block_entity_coord_type
            )
            if entity is not None:
                chunk.entities.append(entity)

        return chunk, palette

    def encode(
        self,
        chunk: "Chunk",
        palette: AnyNDArray,
        max_world_version: VersionIdentifierType,
        box: SelectionBox,
    ) -> MCStructureChunk:
        """
        Take a version-specific chunk and encode it to raw data for the format to store.
        :param chunk: The already translated version-specfic chunk to encode.
        :param palette: The block_palette the ids in the chunk correspond to.
        :type palette: numpy.ndarray[Block]
        :param max_world_version: The key to use to find the encoder.
        :param box: The volume of the chunk to pack.
        :return: Raw data to be stored by the Format.
        """
        entities = []
        for entity in chunk.entities:
            if entity.location in box:
                entities.append(
                    self._encode_entity(
                        entity, self._entity_id_type, self._entity_coord_type
                    ).compound
                )
        block_entities = []
        for block_entity in chunk.block_entities:
            if block_entity.location in box:
                block_entities.append(
                    self._encode_block_entity(
                        block_entity,
                        self._block_entity_id_type,
                        self._block_entity_coord_type,
                    ).compound
                )

        slices = box.create_moved_box(
            (chunk.cx * 16, 0, chunk.cz * 16), subtract=True
        ).slice

        out_palette = numpy.empty(palette.shape, dtype=object)
        for index, block_layers in enumerate(palette):
            blocks_out = []
            for version, block in block_layers:
                block = CompoundTag(
                    {
                        "name": StringTag(f"{block.namespace}:{block.base_name}"),
                        "states": CompoundTag(block.properties),
                    }
                )
                if version:
                    block["version"] = IntTag(version)
                blocks_out.append(block)
            out_palette[index] = blocks_out

        return MCStructureChunk(
            box,
            numpy.asarray(chunk.blocks[slices]),
            out_palette,
            block_entities,
            entities,
        )

    def get_translator(
        self,
        max_world_version: Tuple[PlatformType, VersionNumberTuple],
        data: Any = None,
        translation_manager: "TranslationManager" = None,
    ) -> Tuple["Translator", VersionNumberAny]:
        platform, version_number = max_world_version
        if platform != "bedrock":
            raise ValueError("Platform must be bedrock")
        return Translators.get((platform, version_number)), version_number



from __future__ import annotations

from typing import Union, Iterable, Dict
import time
import numpy
import pickle

from amulet.api.block import Block
from amulet.api.registry import BlockManager
from amulet.api.registry.biome_manager import BiomeManager
from amulet.api.chunk import (
    Biomes,
    BiomesShape,
    Blocks,
    Status,
    BlockEntityDict,
    EntityList,
)
from amulet.api.entity import Entity
from amulet.api.data_types import ChunkCoordinates, VersionIdentifierType
from amulet.api.history.changeable import Changeable


class Chunk(Changeable):
    """
    A class to represent a chunk that exists in a Minecraft world
    """

    def __init__(self, cx: int, cz: int):
        """
        Construct an instance of :class:`Chunk` at the given coordinates.

        :param cx: The x coordinate of the chunk (is not the same as block coordinates)
        :param cz: The z coordinate of the chunk (is not the same as block coordinates)
        """
        super().__init__()
        self._cx, self._cz = cx, cz
        self._changed_time = 0.0

        self._blocks = None
        self.__block_palette = BlockManager()
        self.__biome_palette = BiomeManager()
        self._biomes = None
        self._entities = EntityList()
        self._block_entities = BlockEntityDict()
        self._status = Status()
        self._misc = {}  # all entries that are not important enough to get an attribute

        # TODO: remove these variables. They are temporary until the translator supports entities
        self._native_version: VersionIdentifierType = ("java", 0)
        self._native_entities = EntityList()

    def __repr__(self):
        return f"Chunk({self.cx}, {self.cz}, {repr(self._blocks)}, {repr(self._entities)}, {repr(self._block_entities)})"

    def pickle(self) -> bytes:
        """
        Serialise the data in the chunk using pickle and return the resulting bytes.

        :return: Pickled output.
        """
        chunk_data = (
            self._cx,
            self._cz,
            self._changed_time,
            {sy: self.blocks.get_sub_chunk(sy) for sy in self.blocks.sub_chunks},
            self.biomes.to_raw(),
            self._entities.data,
            tuple(self._block_entities.data.values()),
            self._status.value,
            self.misc,
            self._native_entities,
            self._native_version,
        )
        return pickle.dumps(chunk_data)

    @classmethod
    def unpickle(
        cls,
        pickled_bytes: bytes,
        block_palette: BlockManager,
        biome_palette: BiomeManager,
    ) -> Chunk:
        """
        Deserialise the pickled input and unpack the data into an instance of :class:`Chunk`

        :param pickled_bytes: The bytes returned from :func:`pickle`
        :param block_palette: The instance of :class:`BlockManager` associated with the level.
        :param biome_palette: The instance of :class:`BiomeManager` associated with the level.
        :return: An instance of :class:`Chunk` containing the unpickled data.
        """
        chunk_data = pickle.loads(pickled_bytes)
        self = cls(*chunk_data[:2])
        (
            self.blocks,
            biomes,
            self.entities,
            self.block_entities,
            self.status,
            self.misc,
            self._native_entities,
            self._native_version,
        ) = chunk_data[3:]

        self._biomes = Biomes.from_raw(*biomes)

        self._changed_time = chunk_data[2]
        self._block_palette = block_palette
        self._biome_palette = biome_palette
        self.changed = False
        return self

    @property
    def cx(self) -> int:
        """The chunk's x coordinate"""
        return self._cx

    @property
    def cz(self) -> int:
        """The chunk's z coordinate"""
        return self._cz

    @property
    def coordinates(self) -> ChunkCoordinates:
        """The chunk's x and z coordinates"""
        return self._cx, self._cz

    @property
    def changed(self) -> bool:
        """
        Has the chunk changed since the last undo point. Is used to track which chunks have changed.

        >>> chunk = Chunk(0, 0)
        >>> # Run this to notify that the chunk data has changed.
        >>> chunk.changed = True

        :setter: Set this to ``True`` if you have modified the chunk in any way.
        :return: ``True`` if the chunk has been changed since the last undo point, ``False`` otherwise.
        """
        return self._changed

    @changed.setter
    def changed(self, changed: bool):
        assert isinstance(changed, bool), "Changed value must be a bool"
        self._changed = changed
        if changed:
            self._changed_time = time.time()

    @property
    def changed_time(self) -> float:
        """
        The last time the chunk was changed

        Used to track if the chunk was changed since the last save snapshot and if the chunk model needs rebuilding.
        """
        return self._changed_time

    @property
    def blocks(self) -> Blocks:
        """
        The block array for the chunk.

        This is a custom class that stores a numpy array per sub-chunk.

        The values in the arrays are indexes into :attr:`block_palette`.
        """
        if self._blocks is None:
            self._blocks = Blocks()
        return self._blocks

    @blocks.setter
    def blocks(self, value: Union[Dict[int, numpy.ndarray], Blocks, None]):
        if isinstance(value, dict):
            value: Dict[int, numpy.ndarray]
            value = {k: v.astype(numpy.uint32) for k, v in value.items()}
        self._blocks = Blocks(value)

    def get_block(self, dx: int, y: int, dz: int) -> Block:
        """
        Get the universal Block object at the given location within the chunk.

        :param dx: The x coordinate within the chunk. 0-15 inclusive
        :param y: The y coordinate within the chunk. This can be any integer.
        :param dz: The z coordinate within the chunk. 0-15 inclusive
        :return: The universal Block object representation of the block at that location
        """
        return self.block_palette[self.blocks[dx, y, dz]]

    def set_block(self, dx: int, y: int, dz: int, block: Block):
        """
        Set the universal Block object at the given location within the chunk.

        :param dx: The x coordinate within the chunk. 0-15 inclusive
        :param y: The y coordinate within the chunk. This can be any integer.
        :param dz: The z coordinate within the chunk. 0-15 inclusive
        :param block: The universal Block object to set at the given location
        """
        self.blocks[dx, y, dz] = self.block_palette.get_add_block(block)

    @property
    def _block_palette(self) -> BlockManager:
        """The block block_palette for the chunk.
        Usually will refer to a global block block_palette."""
        return self.__block_palette

    @_block_palette.setter
    def _block_palette(self, new_block_palette: BlockManager):
        """Change the block block_palette for the chunk.
        This will change the block block_palette but leave the block array unchanged.
        Only use this if you know what you are doing.
        Designed for internal use. You probably want to use Chunk.block_palette"""
        assert isinstance(new_block_palette, BlockManager)
        self.__block_palette = new_block_palette

    @property
    def block_palette(self) -> BlockManager:
        """
        The block block_palette for the chunk.

        Usually will refer to the level's global block_palette.
        """
        return self._block_palette

    @block_palette.setter
    def block_palette(self, new_block_palette: BlockManager):
        """
        Change the block block_palette for the chunk.

        This will copy over all block states from the old block_palette and remap the block indexes to use the new block_palette.
        """
        assert isinstance(new_block_palette, BlockManager)
        if new_block_palette is not self._block_palette:
            # if current block block_palette and the new block block_palette are not the same object
            if self._block_palette:
                # if there are blocks in the current block block_palette remap the data
                block_lut = numpy.array(
                    [
                        new_block_palette.get_add_block(block)
                        for block in self._block_palette.blocks
                    ],
                    dtype=numpy.uint32,
                )
                for cy in self.blocks.sub_chunks:
                    self.blocks.add_sub_chunk(
                        cy, block_lut[self.blocks.get_sub_chunk(cy)]
                    )

            self.__block_palette = new_block_palette

    @property
    def biomes(self) -> Biomes:
        """
        The biome array for the chunk.

        This is a custom class that stores numpy arrays. See the :class:`Biomes` documentation for more information.

        The values in the arrays are indexes into :attr:`biome_palette`.
        """
        if self._biomes is None:
            self._biomes = Biomes()
        return self._biomes

    @biomes.setter
    def biomes(self, value: Union[Biomes, Dict[int, numpy.ndarray]]):
        self._biomes = Biomes(value)

    @property
    def _biome_palette(self) -> BiomeManager:
        """
        The biome_palette for the chunk.

        Usually will refer to a global biome_palette.
        """
        return self.__biome_palette

    @_biome_palette.setter
    def _biome_palette(self, new_biome_palette: BiomeManager):
        """
        Change the biome_palette for the chunk.

        This will change the biome_palette but leave the biome array unchanged.

        Only use this if you know what you are doing.

        Designed for internal use. You probably want to use Chunk.biome_palette
        """
        assert isinstance(new_biome_palette, BiomeManager)
        self.__biome_palette = new_biome_palette

    @property
    def biome_palette(self) -> BiomeManager:
        """
        The biome block_palette for the chunk.

        Usually will refer to the level's global biome_palette.
        """
        return self._biome_palette

    @biome_palette.setter
    def biome_palette(self, new_biome_palette: BiomeManager):
        """
        Change the biome_palette for the chunk.

        This will copy over all biome states from the old biome_palette and remap the biome indexes to use the new_biome_palette.
        """
        assert isinstance(new_biome_palette, BiomeManager)
        if new_biome_palette is not self._biome_palette:
            # if current biome_palette and the new biome_palette are not the same object
            if self._biome_palette:
                # if there are biomes in the current biome_palette remap the data
                biome_lut = numpy.array(
                    [
                        new_biome_palette.get_add_biome(biome)
                        for biome in self._biome_palette.biomes
                    ],
                    dtype=numpy.uint32,
                )
                if self.biomes.dimension == BiomesShape.Shape2D:
                    self.biomes = biome_lut[self.biomes]
                elif self.biomes.dimension == BiomesShape.Shape3D:
                    self.biomes = {
                        sy: biome_lut[self.biomes.get_section(sy)]
                        for sy in self.biomes.sections
                    }

            self.__biome_palette = new_biome_palette

    @property
    def entities(self) -> EntityList:
        """
        Property that returns the chunk's entity list. Setting this property replaces the chunk's entity list

        :return: A list of all the entities contained in the chunk
        """
        return self._entities

    @entities.setter
    def entities(self, value: Iterable[Entity]):
        """
        :param value: The new entity list
        :type value: list
        :return:
        """
        if self._entities != value:
            self._entities = EntityList(value)

    @property
    def block_entities(self) -> BlockEntityDict:
        """
        Property that returns the chunk's block entity list. Setting this property replaces the chunk's block entity list

        :return: A list of all the block entities contained in the chunk
        """
        return self._block_entities

    @block_entities.setter
    def block_entities(self, value: BlockEntityDict.InputType):
        """
        :param value: The new block entity list
        :type value: list
        :return:
        """
        if self._block_entities != value:
            self._block_entities = BlockEntityDict(value)

    @property
    def status(self) -> Status:
        """
        A class containing the chunk's generation status.
        """
        return self._status

    @status.setter
    def status(self, value: Union[float, int, str]):
        self._status.value = value

    @property
    def misc(self) -> dict:
        """
        Extra data that exists in a chunk but does not have its own location.

        Data in here is not guaranteed to exist and may get moved at a later date.
        """
        return self._misc

    @misc.setter
    def misc(self, misc: dict):
        assert isinstance(misc, dict), "misc must be a dictionary."
        self._misc = misc



from __future__ import annotations

import struct
from typing import (
    Dict,
    Set,
    Optional,
    List,
    TYPE_CHECKING,
    Tuple,
    Union,
)
from threading import RLock
import logging
from contextlib import suppress

from amulet_nbt import (
    NamedTag,
    LongTag,
    StringTag,
    CompoundTag,
    NBTLoadError,
    load as load_nbt,
    utf8_escape_decoder,
    utf8_escape_encoder,
)

from amulet.api.errors import ChunkDoesNotExist, DimensionDoesNotExist
from amulet.api.data_types import ChunkCoordinates
from leveldb import LevelDB
from .chunk import ChunkData

if TYPE_CHECKING:
    from amulet.api.data_types import Dimension
    from .format import LevelDBFormat

log = logging.getLogger(__name__)

InternalDimension = Optional[int]
OVERWORLD = "minecraft:overworld"
THE_NETHER = "minecraft:the_nether"
THE_END = "minecraft:the_end"


class ActorCounter:
    _lock: RLock
    _session: int
    _count: int

    def __init__(self):
        self._lock = RLock()
        self._session = -1
        self._count = 0

    @classmethod
    def from_level(cls, level: LevelDBFormat):
        session = level.root_tag.compound.get_long(
            "worldStartCount", LongTag(0xFFFFFFFF)
        ).py_int
        # for some reason this is a signed int stored in a signed long. Manually apply the sign correctly
        session -= (session & 0x80000000) << 1

        # create the counter object and set the session
        counter = ActorCounter()
        counter._session = session

        # increment and write back so there are no conflicts
        session -= 1
        if session < 0:
            session += 0x100000000
        level.root_tag.compound["worldStartCount"] = LongTag(session)
        level.root_tag.save()

        return counter

    def next(self) -> Tuple[int, int]:
        """
        Get the next unique session id and actor counter.
        Session id is usually negative

        :return: Tuple[session id, actor id]
        """
        with self._lock:
            count = self._count
            self._count += 1
        return self._session, count


class LevelDBDimensionManager:
    # tag_ids = {45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 118}

    # A borrowed reference to the leveldb
    _db: LevelDB
    # A class to keep track of unique actor ids
    _actor_counter: Optional[ActorCounter]

    def __init__(self, level: LevelDBFormat):
        """
        :param level: The leveldb format to read data from
        """
        self._db = level.level_db
        self._actor_counter = ActorCounter.from_level(level)
        # self._levels format Dict[level, Dict[Tuple[cx, cz], List[Tuple[full_key, key_extension]]]]
        self._levels: Dict[InternalDimension, Set[ChunkCoordinates]] = {}
        self._dimension_name_map: Dict["Dimension", InternalDimension] = {}
        self._lock = RLock()

        self.register_dimension(None, OVERWORLD)
        self.register_dimension(1, THE_NETHER)
        self.register_dimension(2, THE_END)

        for key in self._db.keys():
            if 9 <= len(key) <= 10 and key[8] in [44, 118]:  # "," "v"
                self._add_chunk(key)

            elif 13 <= len(key) <= 14 and key[12] in [44, 118]:  # "," "v"
                self._add_chunk(key, has_level=True)

    @property
    def dimensions(self) -> List["Dimension"]:
        """A list of all the levels contained in the world"""
        return list(self._dimension_name_map.keys())

    def register_dimension(
        self,
        dimension_internal: InternalDimension,
        dimension_name: Optional["Dimension"] = None,
    ):
        """
        Register a new dimension.

        :param dimension_internal: The internal representation of the dimension
        :param dimension_name: The name of the dimension shown to the user
        :return:
        """
        if dimension_name is None:
            dimension_name: "Dimension" = f"DIM{dimension_internal}"

        with self._lock:
            if (
                dimension_internal not in self._levels
                and dimension_name not in self._dimension_name_map
            ):
                self._levels[dimension_internal] = set()
                self._dimension_name_map[dimension_name] = dimension_internal

    def _get_internal_dimension(self, dimension: "Dimension") -> InternalDimension:
        if dimension in self._dimension_name_map:
            return self._dimension_name_map[dimension]
        else:
            raise DimensionDoesNotExist(dimension)

    def all_chunk_coords(self, dimension: "Dimension") -> Set[ChunkCoordinates]:
        internal_dimension = self._get_internal_dimension(dimension)
        if internal_dimension in self._levels:
            return self._levels[internal_dimension]
        else:
            return set()

    @staticmethod
    def _get_key(cx: int, cz: int, internal_dimension: InternalDimension) -> bytes:
        if internal_dimension is None:
            return struct.pack("<ii", cx, cz)
        else:
            return struct.pack("<iii", cx, cz, internal_dimension)

    def _has_chunk(
        self, cx: int, cz: int, internal_dimension: InternalDimension
    ) -> bool:
        return (
            internal_dimension in self._levels
            and (cx, cz) in self._levels[internal_dimension]
        )

    def has_chunk(self, cx: int, cz: int, dimension: Dimension) -> bool:
        return self._has_chunk(cx, cz, self._get_internal_dimension(dimension))

    def _add_chunk(self, key_: bytes, has_level: bool = False):
        if has_level:
            cx, cz, level = struct.unpack("<iii", key_[:12])
        else:
            cx, cz = struct.unpack("<ii", key_[:8])
            level = None
        if level not in self._levels:
            self.register_dimension(level)
        self._levels[level].add((cx, cz))

    def get_chunk_data(self, cx: int, cz: int, dimension: "Dimension") -> ChunkData:
        """Get a dictionary of chunk key extension in bytes to the raw data in the key.
        chunk key extension are the character(s) after <cx><cz>[level] in the key
        Will raise ChunkDoesNotExist if the chunk does not exist
        """
        internal_dimension = self._get_internal_dimension(dimension)
        if self._has_chunk(cx, cz, internal_dimension):
            prefix = self._get_key(cx, cz, internal_dimension)
            prefix_len = len(prefix)
            iter_end = prefix + b"\xff\xff\xff\xff"

            chunk_data = ChunkData()
            for key, val in self._db.iterate(prefix, iter_end):
                if key[:prefix_len] == prefix and len(key) <= prefix_len + 2:
                    chunk_data[key[prefix_len:]] = val

            with suppress(KeyError):
                digp_key = b"digp" + prefix
                digp = self._db.get(digp_key)
                chunk_data[
                    b"digp"
                ] = b""  # The presence of this key signals to the put method that this should be created and written
                for i in range(0, (len(digp) // 8) * 8, 8):
                    actor_key = b"actorprefix" + digp[i : i + 8]
                    try:
                        actor_bytes = self._db.get(actor_key)
                        actor = load_nbt(
                            actor_bytes,
                            little_endian=True,
                            string_decoder=utf8_escape_decoder,
                        )
                        actor_tag = actor.compound
                    except KeyError:
                        log.error(f"Could not find actor {actor_key}. Skipping.")
                    except NBTLoadError:
                        log.error(f"Failed to parse actor {actor_key}. Skipping.")
                    else:
                        actor_tag.pop("UniqueID", None)
                        internal_components = actor_tag.setdefault(
                            "internalComponents",
                            CompoundTag(
                                EntityStorageKeyComponent=CompoundTag(
                                    StorageKey=StringTag()
                                )
                            ),
                        )  # 717
                        if (
                            isinstance(internal_components, CompoundTag)
                            and internal_components
                        ):
                            if "EntityStorageKeyComponent" in internal_components:
                                # it is an entity
                                entity_component = internal_components[
                                    "EntityStorageKeyComponent"
                                ]
                                if isinstance(entity_component, CompoundTag):
                                    # delete the storage key component
                                    if isinstance(
                                        entity_component.get("StorageKey"), StringTag
                                    ):
                                        del entity_component["StorageKey"]
                                    # if there is no other data then delete internalComponents
                                    if (
                                        len(entity_component) == 0
                                        and len(internal_components) == 1
                                    ):
                                        del actor_tag["internalComponents"]
                                    else:
                                        log.warning(
                                            f"Extra components found {repr(entity_component)}"
                                        )
                                else:
                                    log.warning(
                                        f"Unrecognised EntityStorageKeyComponent type {repr(entity_component)}"
                                    )

                                chunk_data.entity_actor.append(actor)
                            else:
                                # it is an unknown actor
                                log.warning(
                                    f"Actor {actor_key} has an unknown format. Please report this to a developer {repr(internal_components)}"
                                )
                                for k, v in internal_components.items():
                                    if isinstance(v, CompoundTag) and isinstance(
                                        v.get("StorageKey"), StringTag
                                    ):
                                        v["StorageKey"] = StringTag()
                                chunk_data.unknown_actor.append(actor)
                        else:
                            log.error(
                                f"internalComponents was not valid for actor {actor_key}. Skipping."
                            )
                            continue

            return chunk_data
        else:
            raise ChunkDoesNotExist

    def put_chunk_data(
        self,
        cx: int,
        cz: int,
        chunk_data: ChunkData,
        dimension: "Dimension",
    ):
        """pass data to the region file class"""
        # get the region key
        internal_dimension = self._get_internal_dimension(dimension)
        self._levels[internal_dimension].add((cx, cz))
        key_prefix = self._get_key(cx, cz, internal_dimension)

        batch = {}

        if b"digp" in chunk_data:
            # if writing the digp key we need to delete all actors pointed to by the old digp key otherwise there will be memory leaks
            digp_key = b"digp" + key_prefix
            try:
                old_digp = self._db.get(digp_key)
            except KeyError:
                pass
            else:
                for i in range(0, len(old_digp) // 8 * 8, 8):
                    actor_key = b"actorprefix" + old_digp[i : i + 8]
                    self._db.delete(actor_key)

            digp = []

            def add_actor(actor: NamedTag, is_entity: bool):
                if not (
                    isinstance(actor, NamedTag) and isinstance(actor.tag, CompoundTag)
                ):
                    log.error(f"Actor must be a NamedTag[Compound]")
                    return
                actor_tag = actor.compound
                internal_components = actor_tag.setdefault(
                    "internalComponents", CompoundTag()
                )
                if not isinstance(internal_components, CompoundTag):
                    log.error(
                        f"Invalid internalComponents value. Must be Compound. Skipping. {repr(internal_components)}"
                    )
                    return

                if is_entity:
                    entity_storage = internal_components.setdefault(
                        "EntityStorageKeyComponent", CompoundTag()
                    )
                    if not isinstance(entity_storage, CompoundTag):
                        log.error(
                            f"Invalid EntityStorageKeyComponent value. Must be Compound. Skipping. {repr(entity_storage)}"
                        )
                        return
                    storages = [entity_storage]
                else:
                    storages = []
                    for storage in internal_components.values():
                        if (
                            isinstance(storage, CompoundTag)
                            and storage.get("StorageKey") == StringTag()
                        ):
                            storages.append(storage)
                    if not storages:
                        log.error(
                            f"No valid StorageKeyComponent to write in for unknown actor {internal_components}"
                        )
                        return

                session, uid = self._actor_counter.next()
                # session is already negative
                key = struct.pack(">ii", -session, uid)
                # b'\x00\x00\x00\x01\x00\x00\x00\x0c' 1, 12
                for storage in storages:
                    storage["StorageKey"] = StringTag(utf8_escape_decoder(key))
                # -4294967284 ">q" b'\xff\xff\xff\xff\x00\x00\x00\x0c' ">ii" -1, 12
                actor_tag["UniqueID"] = LongTag(
                    struct.unpack(">q", struct.pack(">ii", session, uid))[0]
                )

                batch[b"actorprefix" + key] = actor.save_to(
                    little_endian=True,
                    compressed=False,
                    string_encoder=utf8_escape_encoder,
                )
                digp.append(key)

            for actor_ in chunk_data.entity_actor:
                add_actor(actor_, True)

            for actor_ in chunk_data.unknown_actor:
                add_actor(actor_, False)

            del chunk_data[b"digp"]
            batch[digp_key] = b"".join(digp)

        for key, val in chunk_data.items():
            key = key_prefix + key
            if val is None:
                self._db.delete(key)
            else:
                batch[key] = val
        if batch:
            self._db.putBatch(batch)

    def delete_chunk(self, cx: int, cz: int, dimension: "Dimension"):
        if dimension not in self._dimension_name_map:
            return  # dimension does not exists so chunk cannot

        internal_dimension = self._dimension_name_map[dimension]
        if not self._has_chunk(cx, cz, internal_dimension):
            return  # chunk does not exists

        prefix = self._get_key(cx, cz, internal_dimension)
        prefix_len = len(prefix)
        iter_end = prefix + b"\xff\xff\xff\xff"
        keys = []
        for key, _ in self._db.iterate(prefix, iter_end):
            if key[:prefix_len] == prefix and len(key) <= prefix_len + 2:
                keys.append(key)

        try:
            digp = self._db.get(b"digp" + prefix)
        except KeyError:
            pass
        else:
            self._db.delete(b"digp" + prefix)
            for i in range(0, len(digp) // 8 * 8, 8):
                actor_key = b"actorprefix" + digp[i : i + 8]
                self._db.delete(actor_key)

        self._levels[internal_dimension].remove((cx, cz))
        for key in keys:
            self._db.delete(key)



from __future__ import annotations

from typing import TYPE_CHECKING

from amulet.api.chunk import Chunk
from amulet.api.data_types import (
    VersionIdentifierType,
)

if TYPE_CHECKING:
    from PyMCTranslate import Version, TranslationManager

from .java_blockstate_translator import JavaBlockstateTranslator


class Java118Translator(JavaBlockstateTranslator):
    # TODO: move this logic into the interface

    @staticmethod
    def is_valid(key):
        return key[0] == "java" and 2836 <= key[1]

    @staticmethod
    def _unpack_biomes(
        translation_manager: "TranslationManager",
        version_identifier: VersionIdentifierType,
        chunk: Chunk,
    ):
        pass

    @staticmethod
    def _pack_biomes(
        translation_manager: "TranslationManager",
        version_identifier: VersionIdentifierType,
        chunk: Chunk,
    ):
        pass


export = Java118Translator



import os
import logging

from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions


entity_support = False

IMG_DIRECTORY = os.path.join(os.path.dirname(__file__), "img")

from .api import *
from amulet.level.load import load_level, load_format

# init a default logger
logging.basicConfig(level=logging.INFO, format="%(levelname)s - %(message)s")



from amulet.api.selection import SelectionGroup
from amulet.api.data_types import Dimension
from amulet.api.chunk import Chunk
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from amulet.api.level import BaseLevel


def delete_chunk(
    world: "BaseLevel",
    dimension: Dimension,
    source_box: SelectionGroup,
    load_original: bool = True,
):
    chunks = [
        (cx, cz)
        for (cx, cz) in source_box.chunk_locations()
        if world.has_chunk(cx, cz, dimension)
    ]
    iter_count = len(chunks)
    for count, (cx, cz) in enumerate(chunks):
        world.delete_chunk(cx, cz, dimension)

        if not load_original:
            # this part is kind of hacky.
            # Work out a propery API to do this.
            key = dimension, cx, cz
            if key not in world.chunks._history_database:
                world.chunks._register_original_entry(key, Chunk(cx, cz))

        yield (count + 1) / iter_count



class Changeable:
    """A class to track if an object has been changed."""

    def __init__(self):
        self._changed = False

    @property
    def changed(self) -> bool:
        """Has the object been modified since the last revision."""
        return self._changed

    @changed.setter
    def changed(self, changed: bool):
        assert isinstance(changed, bool), "Changed must be a bool"
        self._changed = changed



from __future__ import annotations

import os
import struct
from typing import (
    Tuple,
    Dict,
    Generator,
    Optional,
    List,
    Union,
    Iterable,
    BinaryIO,
    Any,
)
import time
import glob
import shutil
import json
import logging

import portalocker

from amulet_nbt import (
    IntTag,
    LongTag,
    DoubleTag,
    StringTag,
    ListTag,
    CompoundTag,
    NamedTag,
    load as load_nbt,
)
from amulet.api.player import Player, LOCAL_PLAYER
from amulet.api.chunk import Chunk
from amulet.api.selection import SelectionGroup, SelectionBox
from amulet.api.wrapper import WorldFormatWrapper, DefaultSelection
from amulet.utils.format_utils import check_all_exist
from amulet.api.errors import (
    DimensionDoesNotExist,
    ObjectWriteError,
    ChunkLoadError,
    ChunkDoesNotExist,
    PlayerDoesNotExist,
)
from amulet.api.data_types import (
    ChunkCoordinates,
    VersionNumberInt,
    PlatformType,
    DimensionCoordinates,
    AnyNDArray,
    Dimension,
)
from .dimension import AnvilDimensionManager, ChunkDataType
from amulet.api import level as api_level
from amulet.level.interfaces.chunk.anvil.base_anvil_interface import BaseAnvilInterface
from .data_pack import DataPack, DataPackManager

log = logging.getLogger(__name__)

InternalDimension = str
OVERWORLD = "minecraft:overworld"
THE_NETHER = "minecraft:the_nether"
THE_END = "minecraft:the_end"


class AnvilFormat(WorldFormatWrapper[VersionNumberInt]):
    """
    This FormatWrapper class exists to interface with the Java world format.
    """

    def __init__(self, path: str):
        """
        Construct a new instance of :class:`AnvilFormat`.

        This should not be used directly. You should instead use :func:`amulet.load_format`.

        :param path: The file path to the serialised data.
        """
        super().__init__(path)
        self._platform = "java"
        self._root_tag: NamedTag = NamedTag()
        self._levels: Dict[InternalDimension, AnvilDimensionManager] = {}
        self._dimension_name_map: Dict[Dimension, InternalDimension] = {}
        self._mcc_support: Optional[bool] = None
        self._lock_time: Optional[bytes] = None
        self._lock: Optional[BinaryIO] = None
        self._data_pack: Optional[DataPackManager] = None
        self._shallow_load()

    def __del__(self):
        self.close()

    def _shallow_load(self):
        try:
            self._load_level_dat()
        except:
            pass

    def _load_level_dat(self):
        """Load the level.dat file and check the image file"""
        if os.path.isfile(os.path.join(self.path, "icon.png")):
            self._world_image_path = os.path.join(self.path, "icon.png")
        else:
            self._world_image_path = self._missing_world_icon
        self.root_tag = load_nbt(os.path.join(self.path, "level.dat"))

    @staticmethod
    def is_valid(path: str) -> bool:
        if not check_all_exist(path, "level.dat"):
            return False

        try:
            level_dat_root = load_nbt(os.path.join(path, "level.dat")).compound
        except:
            return False

        return "Data" in level_dat_root and "FML" not in level_dat_root

    @property
    def valid_formats(self) -> Dict[PlatformType, Tuple[bool, bool]]:
        return {"java": (True, True)}

    @property
    def version(self) -> VersionNumberInt:
        """The data version number that the world was last opened in. eg 2578"""
        if self._version is None:
            self._version = self._get_version()
        return self._version

    def _get_version(self) -> VersionNumberInt:
        return (
            self.root_tag.compound.get_compound("Data", CompoundTag())
            .get_int("DataVersion", IntTag(-1))
            .py_int
        )

    @property
    def root_tag(self) -> NamedTag:
        """The level.dat data for the level."""
        return self._root_tag

    @root_tag.setter
    def root_tag(self, root_tag: Union[NamedTag, CompoundTag]):
        if isinstance(root_tag, CompoundTag):
            self._root_tag = NamedTag(root_tag)
        elif isinstance(root_tag, NamedTag):
            self._root_tag = root_tag
        else:
            raise ValueError("root_tag must be a CompoundTag or NamedTag")

    @property
    def level_name(self) -> str:
        return (
            self.root_tag.compound.get_compound("Data").get_string("LevelName").py_str
        )

    @level_name.setter
    def level_name(self, value: str):
        self.root_tag.compound.setdefault_compound("Data")["LevelName"] = StringTag(
            value
        )

    @property
    def last_played(self) -> int:
        return (
            self.root_tag.compound.get_compound("Data").get_long("LastPlayed").py_int
            // 1000
        )

    @property
    def game_version_string(self) -> str:
        try:
            return f'Java {self.root_tag.compound.get_compound("Data").get_compound("Version").get_string("Name").py_str}'
        except Exception:
            return f"Java Unknown Version"

    @property
    def data_pack(self) -> DataPackManager:
        if self._data_pack is None:
            packs = []
            enabled_packs = (
                self.root_tag.compound.get_compound("Data")
                .get_compound("DataPacks", CompoundTag())
                .get_list("Enabled", ListTag())
            )
            for pack in enabled_packs:
                if isinstance(pack, StringTag):
                    pack_name: str = pack.py_str
                    if pack_name == "vanilla":
                        pass
                    elif pack_name.startswith("file/"):
                        path = os.path.join(self.path, "datapacks", pack_name[5:])
                        if DataPack.is_path_valid(path):
                            packs.append(DataPack(path))
            self._data_pack = DataPackManager(packs)
        return self._data_pack

    @property
    def dimensions(self) -> List[Dimension]:
        return list(self._dimension_name_map.keys())

    def _register_dimension(
        self,
        relative_dimension_path: InternalDimension,
        dimension_name: Optional[Dimension] = None,
    ):
        """
        Register a new dimension.

        :param relative_dimension_path: The relative path to the dimension directory from the world root. "" for the world root.
        :param dimension_name: The name of the dimension shown to the user
        """
        if dimension_name is None:
            dimension_name: Dimension = relative_dimension_path

        if relative_dimension_path:
            path = os.path.join(self.path, relative_dimension_path)
        else:
            path = self.path

        if (
            relative_dimension_path not in self._levels
            and dimension_name not in self._dimension_name_map
        ):
            self._levels[relative_dimension_path] = AnvilDimensionManager(
                path,
                mcc=self._mcc_support,
                layers=("region",) + ("entities",) * (self.version >= 2681),
            )
            self._dimension_name_map[dimension_name] = relative_dimension_path
            bounds = DefaultSelection
            if self.version >= 2709:  # This number might be smaller
                dimension_settings = (
                    self.root_tag.compound.get_compound("Data", CompoundTag())
                    .get_compound("WorldGenSettings", CompoundTag())
                    .get_compound("dimensions", CompoundTag())
                    .get_compound(dimension_name, CompoundTag())
                )
                dimension_tag = dimension_settings.get("type")
                if isinstance(dimension_tag, StringTag):
                    # the settings are in the data pack
                    dimension_name = dimension_tag.py_str
                    if ":" in dimension_name:
                        namespace, base_name = dimension_name.split(":", 1)
                        if (
                            self.version >= 2834
                            and namespace == "minecraft"
                            and base_name == "overworld"
                        ):
                            bounds = SelectionGroup(
                                SelectionBox(
                                    (-30_000_000, -64, -30_000_000),
                                    (30_000_000, 320, 30_000_000),
                                )
                            )
                        dimension_path = (
                            f"data/{namespace}/dimension_type/{base_name}.json"
                        )
                        if self.data_pack.has_file(dimension_path):
                            with self.data_pack.open(dimension_path) as d:
                                try:
                                    dimension_settings_json = json.load(d)
                                except json.JSONDecodeError:
                                    pass
                                else:
                                    if (
                                        "min_y" in dimension_settings_json
                                        and isinstance(
                                            dimension_settings_json["min_y"], int
                                        )
                                    ):
                                        min_y = dimension_settings_json["min_y"]
                                        if min_y % 16:
                                            min_y = 16 * (min_y // 16)
                                    else:
                                        min_y = 0
                                    if (
                                        "height" in dimension_settings_json
                                        and isinstance(
                                            dimension_settings_json["height"], int
                                        )
                                    ):
                                        height = dimension_settings_json["height"]
                                        if height % 16:
                                            height = -16 * (-height // 16)
                                    else:
                                        height = 256

                                    bounds = SelectionGroup(
                                        SelectionBox(
                                            (-30_000_000, min_y, -30_000_000),
                                            (30_000_000, min_y + height, 30_000_000),
                                        )
                                    )

                elif isinstance(dimension_tag, CompoundTag):
                    # the settings are here
                    dimension_compound_tag = dimension_tag
                    min_y = (
                        dimension_compound_tag.get_int("min_y", IntTag()).py_int // 16
                    ) * 16
                    height = (
                        -dimension_compound_tag.get_int("height", IntTag(256)).py_int
                        // 16
                    ) * -16
                    bounds = SelectionGroup(
                        SelectionBox(
                            (-30_000_000, min_y, -30_000_000),
                            (30_000_000, min_y + height, 30_000_000),
                        )
                    )
                else:
                    log.error(
                        f"Expected dimension_tag to be a StringTag or CompoundTag. Got {repr(dimension_tag)} for dimension {dimension_name}"
                    )

            self._bounds[dimension_name] = bounds

    def _get_interface(self, raw_chunk_data: Optional[Any] = None) -> "Interface":
        from amulet.level.loader import Interfaces

        key = self._get_interface_key(raw_chunk_data)
        return Interfaces.get(key)

    def _get_interface_key(
        self, raw_chunk_data: Optional[ChunkDataType] = None
    ) -> Tuple[str, int]:
        if raw_chunk_data is None:
            return self.max_world_version
        else:
            return (
                self.platform,
                raw_chunk_data.get("region", NamedTag())
                .compound.get_int("DataVersion", IntTag(-1))
                .py_int,
            )

    def _decode(
        self,
        interface: BaseAnvilInterface,
        dimension: Dimension,
        cx: int,
        cz: int,
        raw_chunk_data: ChunkDataType,
    ) -> Tuple[Chunk, AnyNDArray]:
        bounds = self.bounds(dimension).bounds
        return interface.decode(cx, cz, raw_chunk_data, (bounds[0][1], bounds[1][1]))

    def _encode(
        self,
        interface: BaseAnvilInterface,
        chunk: Chunk,
        dimension: Dimension,
        chunk_palette: AnyNDArray,
    ) -> ChunkDataType:
        bounds = self.bounds(dimension).bounds
        return interface.encode(
            chunk, chunk_palette, self.max_world_version, (bounds[0][1], bounds[1][1])
        )

    def _reload_world(self):
        # reload the level.dat in case it has changed
        self._load_level_dat()

        # create the session.lock file (this has mostly been lifted from MCEdit)
        try:
            # open the file for writing and reading and lock it
            self._lock = open(os.path.join(self.path, "session.lock"), "wb+")
            portalocker.lock(self._lock, portalocker.LockFlags.EXCLUSIVE)

            # write the current time to the file
            self._lock_time = struct.pack(">Q", int(time.time() * 1000))
            self._lock.write(self._lock_time)

            # flush the changes to disk
            self._lock.flush()
            os.fsync(self._lock.fileno())

        except Exception as e:
            self._lock_time = None
            if self._lock is not None:
                self._lock.close()
                self._lock = None

            self._is_open = False
            self._has_lock = False
            raise Exception(
                f"Could not access session.lock. The world may be open somewhere else.\n{e}"
            ) from e

        self._is_open = True
        self._has_lock = True

        # the real number might actually be lower
        self._mcc_support = self.version > 2203

        self._levels.clear()
        self._bounds.clear()

        # load all the levels
        self._register_dimension("", OVERWORLD)
        self._register_dimension("DIM-1", THE_NETHER)
        self._register_dimension("DIM1", THE_END)

        for dir_name in os.listdir(self.path):
            level_path = os.path.join(self.path, dir_name)
            if os.path.isdir(level_path) and dir_name.startswith("DIM"):
                if AnvilDimensionManager.level_regex.fullmatch(dir_name) is None:
                    continue
                self._register_dimension(dir_name)

        for dimension_path in glob.glob(
            os.path.join(glob.escape(self.path), "dimensions", "*", "*", "region")
        ):
            dimension_path_split = dimension_path.split(os.sep)
            dimension_name = f"{dimension_path_split[-3]}:{dimension_path_split[-2]}"
            self._register_dimension(
                os.path.dirname(os.path.relpath(dimension_path, self.path)),
                dimension_name,
            )

    def _open(self):
        """Open the database for reading and writing"""
        self._reload_world()

    def _create(
        self,
        overwrite: bool,
        bounds: Union[
            SelectionGroup, Dict[Dimension, Optional[SelectionGroup]], None
        ] = None,
        **kwargs,
    ):
        if os.path.isdir(self.path):
            if overwrite:
                shutil.rmtree(self.path)
            else:
                raise ObjectWriteError(
                    f"A world already exists at the path {self.path}"
                )
        self._version = self.translation_manager.get_version(
            self.platform, self.version
        ).data_version

        self.root_tag = root = CompoundTag()
        root["Data"] = data = CompoundTag()
        data["version"] = IntTag(19133)
        data["DataVersion"] = IntTag(self._version)
        data["LastPlayed"] = LongTag(int(time.time() * 1000))
        data["LevelName"] = StringTag("World Created By Amulet")

        os.makedirs(self.path, exist_ok=True)
        self.root_tag.save_to(os.path.join(self.path, "level.dat"))
        self._reload_world()

    @property
    def has_lock(self) -> bool:
        if self._has_lock:
            self._lock.seek(0)
            return self._lock.read(8) == self._lock_time
        return False

    def pre_save_operation(
        self, level: api_level.BaseLevel
    ) -> Generator[float, None, bool]:
        changed_chunks = list(level.chunks.changed_chunks())
        height = self._calculate_height(level, changed_chunks)
        try:
            while True:
                yield next(height) / 2
        except StopIteration as e:
            height_changed = e.value

        # light = self._calculate_light(level, changed_chunks)
        # try:
        #     while True:
        #         yield next(light) / 2
        # except StopIteration as e:
        #     light_changed = e.value

        return height_changed  # or light_changed

    @staticmethod
    def _calculate_height(
        level: api_level.BaseLevel, chunks: List[DimensionCoordinates]
    ) -> Generator[float, None, bool]:
        """Calculate the height values for chunks."""
        chunk_count = len(chunks)
        # it looks like the game recalculates the height value if not defined.
        # Just delete the stored height values so that they do not get written back.
        # tested as of 1.12.2. This may not be true for older versions.
        changed = False
        for i, (dimension, cx, cz) in enumerate(chunks):
            try:
                chunk = level.get_chunk(cx, cz, dimension)
            except ChunkLoadError:
                pass
            else:
                changed_ = False
                changed_ |= chunk.misc.pop("height_mapC", None) is not None
                changed_ |= chunk.misc.pop("height_map256IA", None) is not None
                if changed_:
                    changed = True
                    chunk.changed = True
            yield i / chunk_count
        return changed

    @staticmethod
    def _calculate_light(
        level: api_level.BaseLevel, chunks: List[DimensionCoordinates]
    ) -> Generator[float, None, bool]:
        """Calculate the height values for chunks."""
        # this is needed for before 1.14
        chunk_count = len(chunks)
        changed = False
        if level.level_wrapper.version < 1934:
            # the version may be less than 1934 but is at least 1924
            # calculate the light values
            pass
            # TODO
        else:
            # the game will recalculate the light levels
            for i, (dimension, cx, cz) in enumerate(chunks):
                try:
                    chunk = level.get_chunk(cx, cz, dimension)
                except ChunkLoadError:
                    pass
                else:
                    changed_ = False
                    changed_ |= chunk.misc.pop("block_light", None) is not None
                    changed_ |= chunk.misc.pop("sky_light", None) is not None
                    if changed_:
                        changed = True
                        chunk.changed = True
                yield i / chunk_count
        return changed

    def _save(self):
        """Save the data back to the disk database"""
        os.makedirs(self.path, exist_ok=True)
        self.root_tag.save_to(os.path.join(self.path, "level.dat"))
        # TODO: save other world data

    def _close(self):
        """Close the disk database"""
        if self._lock is not None:
            portalocker.unlock(self._lock)
            self._lock.close()

    def unload(self):
        for level in self._levels.values():
            level.unload()

    def _has_dimension(self, dimension: Dimension):
        return (
            dimension in self._dimension_name_map
            and self._dimension_name_map[dimension] in self._levels
        )

    def _get_dimension(self, dimension: Dimension):
        self._verify_has_lock()
        if self._has_dimension(dimension):
            return self._levels[self._dimension_name_map[dimension]]
        else:
            raise DimensionDoesNotExist(dimension)

    def all_chunk_coords(self, dimension: Dimension) -> Iterable[ChunkCoordinates]:
        if self._has_dimension(dimension):
            yield from self._get_dimension(dimension).all_chunk_coords()

    def has_chunk(self, cx: int, cz: int, dimension: Dimension) -> bool:
        return self._has_dimension(dimension) and self._get_dimension(
            dimension
        ).has_chunk(cx, cz)

    def _delete_chunk(self, cx: int, cz: int, dimension: Dimension):
        """Delete a chunk from a given dimension"""
        if self._has_dimension(dimension):
            self._get_dimension(dimension).delete_chunk(cx, cz)

    # TODO: add a new version of this method that handles all the raw data
    def put_raw_chunk_data(
        self, cx: int, cz: int, data: NamedTag, dimension: Dimension
    ):
        """
        Commit the raw chunk data to the FormatWrapper cache.

        Call :meth:`save` to push all the cache data to the level.

        :param cx: The x coordinate of the chunk.
        :param cz: The z coordinate of the chunk.
        :param data: The raw data to commit to the level.
        :param dimension: The dimension to load the data from.
        """
        self._verify_has_lock()
        self._put_raw_chunk_data(cx, cz, {"region": data}, dimension)

    def _put_raw_chunk_data(
        self, cx: int, cz: int, data: ChunkDataType, dimension: Dimension
    ):
        self._get_dimension(dimension).put_chunk_data_layers(cx, cz, data)

    # TODO: add a new version of this method that handles all the raw data
    def get_raw_chunk_data(self, cx: int, cz: int, dimension: Dimension) -> NamedTag:
        """
        Return the raw data as loaded from disk.

        :param cx: The x coordinate of the chunk.
        :param cz: The z coordinate of the chunk.
        :param dimension: The dimension to load the data from.
        :return: The raw chunk data.
        """
        # TODO: make this return layer data
        return self._safe_load(
            self._legacy_get_raw_chunk_data,
            (cx, cz, dimension),
            "Error loading chunk {} {} {}",
            ChunkLoadError,
            ChunkDoesNotExist,
        )

    def _legacy_get_raw_chunk_data(
        self, cx: int, cz: int, dimension: Dimension
    ) -> NamedTag:
        layers = self._get_raw_chunk_data(cx, cz, dimension)
        if "region" in layers:
            return layers["region"]
        else:
            raise ChunkDoesNotExist

    def _get_raw_chunk_data(
        self, cx: int, cz: int, dimension: Dimension
    ) -> ChunkDataType:
        """
        Return the raw data as loaded from disk.

        :param cx: The x coordinate of the chunk.
        :param cz: The z coordinate of the chunk.
        :param dimension: The dimension to load the data from.
        :return: The raw chunk data.
        """
        return self._get_dimension(dimension).get_chunk_data_layers(cx, cz)

    def all_player_ids(self) -> Iterable[str]:
        """
        Returns a generator of all player ids that are present in the level
        """
        for f in glob.iglob(
            os.path.join(glob.escape(self.path), "playerdata", "*.dat")
        ):
            yield os.path.splitext(os.path.basename(f))[0]
        if self.has_player(LOCAL_PLAYER):
            yield LOCAL_PLAYER

    def has_player(self, player_id: str) -> bool:
        if player_id == LOCAL_PLAYER:
            return "Player" in self.root_tag.compound.get_compound("Data")
        else:
            return os.path.isfile(
                os.path.join(self.path, "playerdata", f"{player_id}.dat")
            )

    def _load_player(self, player_id: str) -> Player:
        """
        Gets the :class:`Player` object that belongs to the specified player id

        If no parameter is supplied, the data of the local player will be returned

        :param player_id: The desired player id
        :return: A Player instance
        """
        player_nbt = self._get_raw_player_data(player_id)
        dimension = player_nbt["Dimension"]
        # TODO: rework this when there is better dimension support.
        if isinstance(dimension, IntTag):
            if -1 <= dimension.py_int <= 1:
                dimension_str = {-1: THE_NETHER, 0: OVERWORLD, 1: THE_END}[
                    dimension.py_int
                ]
            else:
                dimension_str = f"DIM{dimension}"
        elif isinstance(dimension, StringTag):
            dimension_str = dimension.py_str
        else:
            dimension_str = OVERWORLD
        if dimension_str not in self._dimension_name_map:
            dimension_str = OVERWORLD

        # get the players position
        pos_data = player_nbt.get("Pos")
        if (
            isinstance(pos_data, ListTag)
            and len(pos_data) == 3
            and pos_data.list_data_type == DoubleTag.tag_id
        ):
            position = tuple(map(float, pos_data))
            position = tuple(
                p if -100_000_000 <= p <= 100_000_000 else 0.0 for p in position
            )
        else:
            position = (0.0, 0.0, 0.0)

        # get the players rotation
        rot_data = player_nbt.get("Rotation")
        if (
            isinstance(rot_data, ListTag)
            and len(rot_data) == 2
            and rot_data.list_data_type == DoubleTag.tag_id
        ):
            rotation = tuple(map(float, rot_data))
            rotation = tuple(
                p if -100_000_000 <= p <= 100_000_000 else 0.0 for p in rotation
            )
        else:
            rotation = (0.0, 0.0)

        return Player(
            player_id,
            dimension_str,
            position,
            rotation,
        )

    def _get_raw_player_data(self, player_id: str) -> CompoundTag:
        if player_id == LOCAL_PLAYER:
            if "Player" in self.root_tag.compound.get_compound("Data"):
                return self.root_tag.compound.get_compound("Data").get_compound(
                    "Player"
                )
            else:
                raise PlayerDoesNotExist("Local player doesn't exist")
        else:
            path = os.path.join(self.path, "playerdata", f"{player_id}.dat")
            if os.path.exists(path):
                return load_nbt(path).compound
            raise PlayerDoesNotExist(f"Player {player_id} does not exist")


if __name__ == "__main__":
    import sys

    world_path = sys.argv[1]
    world = AnvilDimensionManager(world_path)
    chunk_ = world.get_chunk_data(0, 0)
    print(chunk_)
    world.put_chunk_data(0, 0, chunk_)



# A cache for objects implemented using leveldb for speed.
import os
import shutil
import time
import logging
import glob
import re
import tempfile

import portalocker

from amulet.api.paths import get_cache_dir

log = logging.getLogger(__name__)


def _clear_legacy_cache():
    legacy_cache_dir = get_cache_dir()
    world_temp_dir = os.path.join(legacy_cache_dir, "world_temp")
    if not os.path.isdir(world_temp_dir):
        return
    paths = [
        os.path.join(legacy_cache_dir, t)
        for t in os.listdir(world_temp_dir)
        if t.isnumeric() and int(t) < (time.time() - 7 * 24 * 3600)
    ]
    if paths:
        log.info("Removing legacy cache.")

        # remove all cache directories that were created before a week ago
        # Sometimes if the program is stopped or it crashes the cleanup won't happen
        for path in paths:
            shutil.rmtree(path, ignore_errors=True)


_clear_legacy_cache()


TempPattern = re.compile(r"amulettmp.*?-(?P<time>\d+)")


def _clear_temp_dirs():
    """
    Try and delete historic temporary directories.
    If things went very wrong in past sessions temporary directories may still exist.
    """
    temp_dir = tempfile.gettempdir()
    for path in glob.glob(os.path.join(glob.escape(temp_dir), "amulettmp*")):
        name = os.path.relpath(path, temp_dir)
        match = TempPattern.fullmatch(name)
        if match and int(match.group("time")) < (time.time() - 7 * 24 * 3600):
            lock_path = os.path.join(path, "lock")
            if os.path.exists(lock_path):
                with open(lock_path) as lock:
                    # make sure it is not locked by another process
                    try:
                        portalocker.lock(lock, portalocker.LockFlags.EXCLUSIVE)
                    except:
                        continue
                    else:
                        portalocker.unlock(lock)
            shutil.rmtree(path, ignore_errors=True)


_clear_temp_dirs()


class TempDir(str):
    """
    A temporary directory to do with as you wish.

    >>> t = TempDir()
    >>> path = os.path.join(t, "your_file.txt")  # TempDir is a subclass of str
    >>> # make sure all files in the temporary directory are closed before releasing or closing this object.
    >>> # The temporary directory will be deleted when the last reference to `t` is lost or when `t.close()` is called
    """

    def __new__(cls):
        self = super().__new__(
            cls,
            tempfile.mkdtemp(
                prefix="amulettmp",
                suffix=f"-{time.time():.0f}",
                dir=get_cache_dir(None),
            ),
        )
        self.__lock = open(os.path.join(self, "lock"), "w")
        portalocker.lock(self.__lock, portalocker.LockFlags.EXCLUSIVE)
        return self

    def close(self):
        """Close the lock and delete the directory."""
        if self.__lock is not None:
            portalocker.unlock(self.__lock)
            self.__lock.close()
            self.__lock = None
            shutil.rmtree(self)

    def __del__(self):
        self.close()



from __future__ import annotations

from typing import Tuple, Dict, Optional, TYPE_CHECKING

import numpy

from amulet.utils.world_utils import to_nibble_array
from amulet.api.data_types import (
    AnyNDArray,
    VersionIdentifierTuple,
)
from .leveldb_2 import (
    LevelDB2Interface,
)

if TYPE_CHECKING:
    from amulet.api.chunk.blocks import Blocks


class LevelDB3Interface(LevelDB2Interface):
    chunk_version = 3

    def __init__(self):
        super().__init__()
        self._set_feature("terrain", "2farray")

    def _encode_subchunks(
        self,
        blocks: "Blocks",
        palette: AnyNDArray,
        bounds: Tuple[int, int],
        max_world_version: VersionIdentifierTuple,
    ) -> Dict[int, Optional[bytes]]:
        # encode sub-chunk block format 0
        sections = {}
        palette = numpy.array([b[0][1] for b in palette])
        for cy in range(16):
            if cy in blocks:
                block_sub_array = palette[
                    numpy.transpose(blocks.get_sub_chunk(cy), (0, 2, 1)).ravel()
                ]
                if not numpy.any(block_sub_array):
                    sections[cy] = None
                    continue

                data_sub_array = block_sub_array[:, 1]
                block_sub_array = block_sub_array[:, 0]
                sections[cy] = (
                    b"\00"
                    + block_sub_array.astype("uint8").tobytes()
                    + to_nibble_array(data_sub_array).tobytes()
                )

        return sections


export = LevelDB3Interface



from __future__ import annotations

from typing import Tuple, TYPE_CHECKING

from amulet_nbt import IntTag

import amulet
from .base_anvil_interface import ChunkPathType, ChunkDataType
from .anvil_na import AnvilNAInterface as ParentInterface

if TYPE_CHECKING:
    from amulet.api.chunk import Chunk


class Anvil0Interface(ParentInterface):
    """
    Added the DataVersion tag
    Note that this has not been tested before 1.12
    """

    V = None
    RegionDataVersion: ChunkPathType = (
        "region",
        [("DataVersion", IntTag)],
        IntTag,
    )

    def __init__(self):
        super().__init__()
        self._unregister_decoder(self._decode_v_tag)
        self._unregister_encoder(self._encode_v_tag)

        self._register_post_decoder(self._post_decode_data_version)

    @staticmethod
    def minor_is_valid(key: int):
        return 0 <= key < 1444

    def _post_decode_data_version(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        # all versioned data must get removed from data
        self.get_layer_obj(data, self.RegionDataVersion, pop_last=True)

    def _decode_entities(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        ents = self._decode_entity_list(
            self.get_layer_obj(data, self.Entities, pop_last=True)
        )
        if amulet.entity_support:
            chunk.entities = ents
        else:
            chunk._native_entities.extend(ents)
            chunk._native_version = (
                "java",
                self.get_layer_obj(data, self.RegionDataVersion).py_int,
            )

    def _init_encode(
        self,
        chunk: Chunk,
        max_world_version: Tuple[str, int],
        floor_cy: int,
        height_cy: int,
    ) -> ChunkDataType:
        data = super()._init_encode(chunk, max_world_version, floor_cy, height_cy)
        self.set_layer_obj(data, self.RegionDataVersion, IntTag(max_world_version[1]))
        return data


export = Anvil0Interface



import numpy
from typing import Union, Optional, Dict, Tuple
from copy import deepcopy
from enum import IntEnum


from amulet.api.partial_3d_array import UnboundedPartial3DArray


class Biomes3D(UnboundedPartial3DArray):
    """
    The Biomes3D class is designed to represent a 3D integer array but with no vertical height limit.

    See :class:`UnboundedPartial3DArray` for more information on how this works.
    """

    def __init__(
        self,
        input_array: Optional[Union[Dict[int, numpy.ndarray], "Biomes3D"]] = None,
    ):
        """
        Construct a :class:`Biomes3D` class from the given data.

        :param input_array: Either an instance of :class:`Biomes3D` or a dictionary converting sub-chunk id to a 4x4x4 numpy array.
        """
        if input_array is None:
            input_array = {}
        if isinstance(input_array, Biomes3D):
            input_array: dict = deepcopy(input_array._sections)
        if not isinstance(input_array, dict):
            raise Exception(f"Input array must be Biomes3D or dict, got {input_array}")
        super().__init__(numpy.uint32, 0, (4, 4, 4), (0, 16), sections=input_array)


class BiomesShape(IntEnum):
    """
    An enum of the different states the :class:`Biomes` class can be in.

    >>> ShapeNull = 0  # The biome array does not exist
    >>> Shape2D = 2  # The biome array is a 2D array
    >>> Shape3D = 3  # The biome array is a 3D array
    """

    ShapeNull = 0  # doc: The biome array does not exist
    Shape2D = 2  # doc: The biome array is a 2D array
    Shape3D = 3  # doc: The biome array is a 3D array


class Biomes:
    """
    Biomes is a class used to store a 2D biome array and/or 3D biome array and facilitate switching between them.

    There are three states biomes can be in.

    The first is a blank state where no biomes are defined. This is used when a chunk is partially generated and the biomes array has not be created.

    The second is a 2D array of size (16, 16) used for old Java worlds and Bedrock worlds.

    The last is the 3D state used since Java 1.15. See :class:`Biomes3D` for more information.
    """

    __slots__ = ("_2d", "_3d", "_dimension")
    _2d: Optional[numpy.ndarray]
    _3d: Optional[Biomes3D]
    _dimension: BiomesShape

    def __init__(
        self, array: Union[numpy.ndarray, Biomes3D, Dict[int, numpy.ndarray]] = None
    ):
        """
        Construct a new instance of :class:`Biomes`

        :param array: The array to initialise with. Can be None or not defined to have the empty state.
        """
        self._2d: Optional[numpy.ndarray] = None
        self._3d: Optional[Biomes3D] = None
        if array is None:
            self._dimension = BiomesShape.ShapeNull
        elif isinstance(array, numpy.ndarray):
            assert array.shape == (
                16,
                16,
            ), "If Biomes is given an ndarray it must be 16x16"
            self._2d = array.copy()
            self._dimension = BiomesShape.Shape2D
        elif isinstance(array, (dict, Biomes3D)):
            self._3d = Biomes3D(array)
            self._dimension = BiomesShape.Shape3D

    def to_raw(
        self,
    ) -> Tuple[
        BiomesShape, Optional[numpy.ndarray], Optional[Dict[int, numpy.ndarray]]
    ]:
        """Don't use this method. Use to pickle data."""
        if self._3d is None:
            sections = None
        else:
            sections = self._3d._sections
        return self._dimension, self._2d, sections

    @classmethod
    def from_raw(
        cls,
        dimension: BiomesShape,
        d2: Optional[numpy.ndarray],
        d3: Optional[Dict[int, numpy.ndarray]],
    ) -> "Biomes":
        """Don't use this method. Use to unpickle data."""
        biomes = cls()
        biomes._dimension = dimension
        biomes._2d = d2
        if d3 is not None:
            biomes._3d = Biomes3D(d3)
        return biomes

    @property
    def dimension(self) -> BiomesShape:
        """
        The number of dimensions the data has

        :attr:`BiomesShape.ShapeNull` when there is no data.
        Will error if you try accessing data from it without converting to one of the other formats.

        :attr:`BiomesShape.Shape2D` when the data is a 2D (16, 16) numpy array.

        :attr:`BiomesShape.Shape3D` when the data is a 3D (4, inf, 4) :class:`Biome3D` array made of sections of size (4, 4, 4).
        """
        return self._dimension

    def convert_to_2d(self):
        """
        Convert the data to the 2D 16x16 format from whatever format it was in.

        If it was in the Null state it will be initialised with the first entry in the :obj:`~amulet.api.registry.biome_manager.BiomeManager`.

        In this mode this class will behave like a numpy array.
        """
        if self._2d is None:
            self._2d = numpy.zeros((16, 16), numpy.uint32)
        if self._dimension != BiomesShape.Shape2D:
            if self._dimension == BiomesShape.Shape3D and self._3d is not None:
                # convert from 3D
                self._2d[:, :] = numpy.kron(
                    numpy.reshape(self._3d[:, 0, :], (4, 4)), numpy.ones((4, 4))
                )
            self._dimension = BiomesShape.Shape2D

    def convert_to_3d(self):
        """
        Convert the data to the 3D (4, inf, 4) format from whatever format it was in.

        If it was in the Null state it will be initialised with the first entry in the :class:`~amulet.api.registry.biome_manager.BiomeManager`.

        In this mode this class will behave like the :class:`Biome3D` class.
        """
        if self._3d is None:
            self._3d = Biomes3D()
        if self._dimension != BiomesShape.Shape3D:
            if self._dimension == BiomesShape.Shape2D and self._2d is not None:
                # convert from 2D
                self._3d[:, 0, :] = self._2d[::4, ::4].reshape(4, 1, 4)
            self._dimension = BiomesShape.Shape3D

    def _get_active(self) -> Union[numpy.ndarray, Biomes3D]:
        if self._dimension == BiomesShape.ShapeNull:
            raise Exception(
                "You are trying to use Biomes but have not picked a format. Use one of the convert methods to specify the format."
            )
        elif self._dimension == BiomesShape.Shape2D:
            self.convert_to_2d()
            return self._2d
        elif self._dimension == BiomesShape.Shape3D:
            self.convert_to_3d()
            return self._3d
        else:
            raise Exception("Dimension is invalid. This shouldn't happen")

    def copy(self):
        return self.__copy__()

    def __copy__(self):
        cls = self.__class__
        result = cls.__new__(cls)
        for k in self.__slots__:
            setattr(result, k, getattr(self, k))
        return result

    def __deepcopy__(self, memodict=None):
        cls = self.__class__
        result = cls.__new__(cls)
        memodict[id(self)] = result
        for k in self.__slots__:
            setattr(result, k, deepcopy(getattr(self, k), memodict))
        return result

    def __getattr__(self, item):
        return self._get_active().__getattribute__(item)

    def __contains__(self, item):
        return self._get_active().__contains__(item)

    def __getitem__(self, item):
        return self._get_active().__getitem__(item)

    def __setitem__(self, key, value):
        self._get_active().__setitem__(key, value)
